# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .api_request_parent_context import ApiRequestParentContext
from .external_parent_context import ExternalParentContext
from .node_parent_context import NodeParentContext
from .prompt_deployment_parent_context import PromptDeploymentParentContext
from .span_link import SpanLink
from .workflow_deployment_parent_context import WorkflowDeploymentParentContext
from .workflow_parent_context import WorkflowParentContext
from .workflow_sandbox_parent_context import WorkflowSandboxParentContext
from .array_vellum_value import ArrayVellumValue
import typing
import datetime as dt
from .execution_vellum_value import ExecutionVellumValue
from .workflow_error import WorkflowError
from .workflow_execution_actual import WorkflowExecutionActual
from .workflow_execution_view_online_eval_metric_result import WorkflowExecutionViewOnlineEvalMetricResult
from .workflow_execution_usage_result import WorkflowExecutionUsageResult
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class SlimWorkflowExecutionRead(UniversalBaseModel):
    span_id: str
    parent_context: typing.Optional[WorkflowDeploymentParentContext] = None
    start: dt.datetime
    end: typing.Optional[dt.datetime] = None
    inputs: typing.List[ExecutionVellumValue]
    outputs: typing.List[ExecutionVellumValue]
    error: typing.Optional[WorkflowError] = None
    latest_actual: typing.Optional[WorkflowExecutionActual] = None
    metric_results: typing.List[WorkflowExecutionViewOnlineEvalMetricResult]
    usage_results: typing.Optional[typing.List[WorkflowExecutionUsageResult]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from .workflow_execution_span import WorkflowExecutionSpan
from .node_execution_span import NodeExecutionSpan

VellumSpan = typing.Union[WorkflowExecutionSpan, NodeExecutionSpan]



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .array_vellum_value import ArrayVellumValue
from .vellum_variable_type import VellumVariableType
import typing
from .vellum_value import VellumValue
from .vellum_variable_extensions import VellumVariableExtensions
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class VellumVariable(UniversalBaseModel):
    id: str
    key: str
    type: VellumVariableType
    required: typing.Optional[bool] = None
    default: typing.Optional[VellumValue] = None
    extensions: typing.Optional[VellumVariableExtensions] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import pytest
from uuid import UUID
from typing import Type

from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes.core.templating_node.node import TemplatingNode
from vellum_ee.workflows.display.nodes.vellum.templating_node import BaseTemplatingNodeDisplay
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def _no_display_class(Node: Type[TemplatingNode]):
    return None


def _display_class_with_node_input_ids_by_name(Node: Type[TemplatingNode]):
    class TemplatingNodeDisplay(BaseTemplatingNodeDisplay[Node]):  # type: ignore[valid-type]
        node_input_ids_by_name = {"foo": UUID("fba6a4d5-835a-4e99-afb7-f6a4aed15110")}

    return TemplatingNodeDisplay


def _display_class_with_node_input_ids_by_name_with_inputs_prefix(Node: Type[TemplatingNode]):
    class TemplatingNodeDisplay(BaseTemplatingNodeDisplay[Node]):  # type: ignore[valid-type]
        node_input_ids_by_name = {"inputs.foo": UUID("fba6a4d5-835a-4e99-afb7-f6a4aed15110")}

    return TemplatingNodeDisplay


@pytest.mark.parametrize(
    ["GetDisplayClass", "expected_input_id"],
    [
        (_no_display_class, "91d982a9-6c41-42ac-aff9-7b623c450a55"),
        (_display_class_with_node_input_ids_by_name, "fba6a4d5-835a-4e99-afb7-f6a4aed15110"),
        (_display_class_with_node_input_ids_by_name_with_inputs_prefix, "fba6a4d5-835a-4e99-afb7-f6a4aed15110"),
    ],
    ids=[
        "no_display_class",
        "display_class_with_node_input_ids_by_name",
        "display_class_with_node_input_ids_by_name_with_inputs_prefix",
    ],
)
def test_serialize_node__templating_node_inputs(GetDisplayClass, expected_input_id):
    # GIVEN a templating node with inputs
    class MyTemplatingNode(TemplatingNode):
        inputs = {"foo": "bar"}

    # AND a workflow with the templating node
    class Workflow(BaseWorkflow):
        graph = MyTemplatingNode

    # AND a display class
    GetDisplayClass(MyTemplatingNode)

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should properly serialize the inputs
    my_templating_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["type"] == "TEMPLATING"
    )

    assert my_templating_node["inputs"] == [
        {
            "id": "b7d971ee-9156-46b2-9bf0-9292875211f8",
            "key": "template",
            "value": {
                "combinator": "OR",
                "rules": [
                    {
                        "type": "CONSTANT_VALUE",
                        "data": {
                            "type": "STRING",
                            "value": "",
                        },
                    }
                ],
            },
        },
        {
            "id": expected_input_id,
            "key": "foo",
            "value": {
                "combinator": "OR",
                "rules": [
                    {
                        "type": "CONSTANT_VALUE",
                        "data": {
                            "type": "STRING",
                            "value": "bar",
                        },
                    }
                ],
            },
        },
    ]



import pytest

from vellum import ExecuteApiResponse, VellumSecret as ClientVellumSecret
from vellum.client.core.api_error import ApiError
from vellum.workflows.constants import APIRequestMethod, AuthorizationType
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes import APINode
from vellum.workflows.types.core import VellumSecret


def test_run_workflow__secrets(vellum_client):
    vellum_client.execute_api.return_value = ExecuteApiResponse(
        status_code=200,
        text='{"status": 200, "data": [1, 2, 3]}',
        json_={"data": [1, 2, 3]},
        headers={"X-Response-Header": "bar"},
    )
    vellum_client._client_wrapper.get_headers.return_value = {"User-Agent": "vellum-ai/1.0.6"}

    class SimpleBaseAPINode(APINode):
        method = APIRequestMethod.POST
        authorization_type = AuthorizationType.BEARER_TOKEN
        url = "https://example.vellum.ai"
        json = {
            "key": "value",
        }
        headers = {
            "X-Test-Header": "foo",
        }
        bearer_token_value = VellumSecret(name="secret")

    node = SimpleBaseAPINode()
    terminal = node.run()

    assert vellum_client.execute_api.call_count == 1
    assert vellum_client.execute_api.call_args.kwargs["url"] == "https://example.vellum.ai"
    assert vellum_client.execute_api.call_args.kwargs["body"] == {"key": "value"}
    headers = vellum_client.execute_api.call_args.kwargs["headers"]
    assert headers["X-Test-Header"] == "foo"
    assert "vellum-ai" in headers.get("User-Agent", "")
    bearer_token = vellum_client.execute_api.call_args.kwargs["bearer_token"]
    assert bearer_token == ClientVellumSecret(name="secret")
    assert terminal.headers == {"X-Response-Header": "bar"}


def test_api_node_raises_error_when_api_call_fails(vellum_client):
    # GIVEN an API call that fails
    vellum_client.execute_api.side_effect = ApiError(status_code=400, body="API Error")
    vellum_client._client_wrapper.get_headers.return_value = {"User-Agent": "vellum-ai/1.0.6"}

    class SimpleAPINode(APINode):
        method = APIRequestMethod.GET
        authorization_type = AuthorizationType.BEARER_TOKEN
        url = "https://example.vellum.ai"
        json = {
            "key": "value",
        }
        headers = {
            "X-Test-Header": "foo",
        }
        bearer_token_value = VellumSecret(name="api_key")

    node = SimpleAPINode()

    # WHEN we run the node
    with pytest.raises(NodeException) as excinfo:
        node.run()

    # THEN an exception should be raised
    assert "Failed to prepare HTTP request" in str(excinfo.value)

    # AND the API call should have been made
    assert vellum_client.execute_api.call_count == 1
    assert vellum_client.execute_api.call_args.kwargs["url"] == "https://example.vellum.ai"
    assert vellum_client.execute_api.call_args.kwargs["body"] == {"key": "value"}
    headers = vellum_client.execute_api.call_args.kwargs["headers"]
    assert headers["X-Test-Header"] == "foo"
    assert "vellum-ai" in headers.get("User-Agent", "")


def test_api_node_defaults_to_get_method(vellum_client):
    # GIVEN a successful API response
    vellum_client.execute_api.return_value = ExecuteApiResponse(
        status_code=200,
        text='{"status": 200, "data": [1, 2, 3]}',
        json_={"data": [1, 2, 3]},
        headers={"X-Response-Header": "bar"},
    )

    # AND an API node without a method specified
    class SimpleAPINodeWithoutMethod(APINode):
        authorization_type = AuthorizationType.BEARER_TOKEN
        url = "https://example.vellum.ai"
        headers = {
            "X-Test-Header": "foo",
        }
        bearer_token_value = VellumSecret(name="secret")

    node = SimpleAPINodeWithoutMethod()

    # WHEN we run the node
    node.run()

    # THEN the API call should be made with GET method
    assert vellum_client.execute_api.call_count == 1
    method = vellum_client.execute_api.call_args.kwargs["method"]
    assert method == APIRequestMethod.GET.value


def test_api_node__detects_client_environment_urls__adds_token(mock_httpx_transport, mock_requests, monkeypatch):
    # GIVEN an API node with a URL pointing back to Vellum
    class SimpleAPINodeToVellum(APINode):
        url = "https://api.vellum.ai"

    # AND a mock request sent to the Vellum API would return a 200
    mock_response = mock_requests.get(
        "https://api.vellum.ai",
        status_code=200,
        json={"data": [1, 2, 3]},
    )

    # AND an api key is set
    monkeypatch.setenv("VELLUM_API_KEY", "vellum-api-key-1234")

    # WHEN we run the node
    node = SimpleAPINodeToVellum()
    node.run()

    # THEN the execute_api method should not have been called
    mock_httpx_transport.handle_request.assert_not_called()

    # AND the vellum API should have been called with the correct headers
    assert mock_response.last_request
    assert mock_response.last_request.headers["X-API-Key"] == "vellum-api-key-1234"


def test_api_node__detects_client_environment_urls__does_not_override_headers(
    mock_httpx_transport, mock_requests, monkeypatch
):
    # GIVEN an API node with a URL pointing back to Vellum
    class SimpleAPINodeToVellum(APINode):
        url = "https://api.vellum.ai"
        headers = {
            "X-API-Key": "vellum-api-key-5678",
        }

    # AND a mock request sent to the Vellum API would return a 200
    mock_response = mock_requests.get(
        "https://api.vellum.ai",
        status_code=200,
        json={"data": [1, 2, 3]},
    )

    # AND an api key is set
    monkeypatch.setenv("VELLUM_API_KEY", "vellum-api-key-1234")

    # WHEN we run the node
    node = SimpleAPINodeToVellum()
    node.run()

    # THEN the execute_api method should not have been called
    mock_httpx_transport.handle_request.assert_not_called()

    # AND the vellum API should have been called with the correct headers
    assert mock_response.last_request
    assert mock_response.last_request.headers["X-API-Key"] == "vellum-api-key-5678"


def test_api_node__detects_client_environment_urls__legacy_does_not_override_headers(
    mock_httpx_transport, mock_requests, monkeypatch
):
    # GIVEN an API node with a URL pointing back to Vellum
    class SimpleAPINodeToVellum(APINode):
        url = "https://api.vellum.ai"
        headers = {
            "X_API_KEY": "vellum-api-key-5678",
        }

    # AND a mock request sent to the Vellum API would return a 200
    mock_response = mock_requests.get(
        "https://api.vellum.ai",
        status_code=200,
        json={"data": [1, 2, 3]},
    )

    # AND an api key is set
    monkeypatch.setenv("VELLUM_API_KEY", "vellum-api-key-1234")

    # WHEN we run the node
    node = SimpleAPINodeToVellum()
    node.run()

    # THEN the execute_api method should not have been called
    mock_httpx_transport.handle_request.assert_not_called()

    # AND the vellum API should have been called with the correct headers
    assert mock_response.last_request
    assert mock_response.last_request.headers["X_API_KEY"] == "vellum-api-key-5678"


def test_api_node_raises_error_for_empty_url():
    class APINodeWithEmptyURL(APINode):
        method = APIRequestMethod.GET
        url = ""

    node = APINodeWithEmptyURL()

    with pytest.raises(NodeException) as excinfo:
        node.run()

    assert excinfo.value.code == WorkflowErrorCode.INVALID_INPUTS
    assert "URL is required and must be a non-empty string" in str(excinfo.value)


def test_api_node_raises_error_for_none_url():
    class APINodeWithNoneURL(APINode):
        method = APIRequestMethod.GET
        url = None  # type: ignore

    node = APINodeWithNoneURL()

    with pytest.raises(NodeException) as excinfo:
        node.run()

    assert excinfo.value.code == WorkflowErrorCode.INVALID_INPUTS
    assert "URL is required and must be a non-empty string" in str(excinfo.value)


def test_api_node_raises_error_for_whitespace_url():
    class APINodeWithWhitespaceURL(APINode):
        method = APIRequestMethod.GET
        url = "   "

    node = APINodeWithWhitespaceURL()

    with pytest.raises(NodeException) as excinfo:
        node.run()

    assert excinfo.value.code == WorkflowErrorCode.INVALID_INPUTS
    assert "URL is required and must be a non-empty string" in str(excinfo.value)


def test_api_node_passes_timeout_to_vellum_client(vellum_client):
    """Test that timeout is passed to vellum client as RequestOptions."""
    vellum_client.execute_api.return_value = ExecuteApiResponse(
        status_code=200,
        text='{"result": "success"}',
        json_={"result": "success"},
        headers={"content-type": "application/json"},
    )

    # GIVEN an API node configured with a timeout value of 25 seconds
    class APINodeWithTimeout(APINode):
        method = APIRequestMethod.GET
        authorization_type = AuthorizationType.BEARER_TOKEN
        url = "https://example.com"
        timeout = 25
        bearer_token_value = VellumSecret(name="secret")

    # WHEN the API node is executed
    node = APINodeWithTimeout()
    node.run()

    # THEN the vellum client should be called exactly once
    assert vellum_client.execute_api.call_count == 1
    call_args = vellum_client.execute_api.call_args
    request_options = call_args.kwargs["request_options"]
    # AND the call should include RequestOptions with the correct timeout
    assert request_options is not None
    assert request_options["timeout_in_seconds"] == 25



import pytest
import enum

from pydantic import BaseModel

from vellum.client.types.chat_history_vellum_value import ChatHistoryVellumValue
from vellum.client.types.chat_message import ChatMessage
from vellum.client.types.error_vellum_value import ErrorVellumValue
from vellum.client.types.json_vellum_value import JsonVellumValue
from vellum.client.types.number_vellum_value import NumberVellumValue
from vellum.client.types.search_result import SearchResult
from vellum.client.types.search_result_document import SearchResultDocument
from vellum.client.types.search_results_vellum_value import SearchResultsVellumValue
from vellum.client.types.string_vellum_value import StringVellumValue
from vellum.client.types.string_vellum_value_request import StringVellumValueRequest
from vellum.client.types.vellum_error import VellumError
from vellum.workflows.errors.types import WorkflowError, WorkflowErrorCode
from vellum.workflows.nodes.displayable.bases.utils import primitive_to_vellum_value, primitive_to_vellum_value_request


class MockEnum(enum.Enum):
    FOO = "foo"


class RandomPydanticModel(BaseModel):
    foo: str


@pytest.mark.parametrize(
    ["value", "expected_output"],
    [
        ("hello", StringVellumValue(value="hello")),
        (MockEnum.FOO, StringVellumValue(value="foo")),
        (1, NumberVellumValue(value=1)),
        (1.0, NumberVellumValue(value=1.0)),
        (
            [ChatMessage(role="USER", text="hello")],
            ChatHistoryVellumValue(value=[ChatMessage(role="USER", text="hello")]),
        ),
        (
            [
                SearchResult(
                    text="Search query",
                    score="0.0",
                    keywords=["keywords"],
                    document=SearchResultDocument(label="label"),
                )
            ],
            SearchResultsVellumValue(
                value=[
                    SearchResult(
                        text="Search query",
                        score="0.0",
                        keywords=["keywords"],
                        document=SearchResultDocument(label="label"),
                    )
                ]
            ),
        ),
        (StringVellumValue(value="hello"), StringVellumValue(value="hello")),
        (StringVellumValueRequest(value="hello"), StringVellumValueRequest(value="hello")),
        ({"foo": "bar"}, JsonVellumValue(value={"foo": "bar"})),
        (
            VellumError(message="hello", code="USER_DEFINED_ERROR"),
            ErrorVellumValue(value=VellumError(message="hello", code="USER_DEFINED_ERROR")),
        ),
        (
            WorkflowError(message="hello", code=WorkflowErrorCode.USER_DEFINED_ERROR),
            ErrorVellumValue(value=VellumError(message="hello", code="USER_DEFINED_ERROR")),
        ),
        (RandomPydanticModel(foo="bar"), JsonVellumValue(value={"foo": "bar"})),
    ],
)
def test_primitive_to_vellum_value(value, expected_output):
    assert primitive_to_vellum_value(value) == expected_output


def test_primitive_to_vellum_value_request():
    assert primitive_to_vellum_value_request("hello") == StringVellumValueRequest(value="hello")



# This file was auto-generated by Fern from our API Definition.

from ...core.client_wrapper import SyncClientWrapper
from .raw_client import RawOrganizationsClient
import typing
from ...core.request_options import RequestOptions
from ...types.organization_read import OrganizationRead
from ...core.client_wrapper import AsyncClientWrapper
from .raw_client import AsyncRawOrganizationsClient


class OrganizationsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawOrganizationsClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawOrganizationsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawOrganizationsClient
        """
        return self._raw_client

    def organization_identity(self, *, request_options: typing.Optional[RequestOptions] = None) -> OrganizationRead:
        """
        Retrieves information about the active Organization

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.organizations.organization_identity()
        """
        response = self._raw_client.organization_identity(request_options=request_options)
        return response.data


class AsyncOrganizationsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawOrganizationsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawOrganizationsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawOrganizationsClient
        """
        return self._raw_client

    async def organization_identity(
        self, *, request_options: typing.Optional[RequestOptions] = None
    ) -> OrganizationRead:
        """
        Retrieves information about the active Organization

        Parameters
        ----------
        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        OrganizationRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.organizations.organization_identity()


        asyncio.run(main())
        """
        response = await self._raw_client.organization_identity(request_options=request_options)
        return response.data



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .chat_message import ChatMessage
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class NodeInputCompiledChatHistoryValue(UniversalBaseModel):
    node_input_id: str
    key: str
    type: typing.Literal["CHAT_HISTORY"] = "CHAT_HISTORY"
    value: typing.Optional[typing.List[ChatMessage]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .components_schemas_prompt_version_build_config_sandbox import ComponentsSchemasPromptVersionBuildConfigSandbox
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import typing


class PromptDeploymentReleasePromptVersion(UniversalBaseModel):
    id: str
    build_config: ComponentsSchemasPromptVersionBuildConfigSandbox = pydantic.Field()
    """
    Configuration used to build this prompt version.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



from uuid import UUID
from typing import Callable, Dict, Generic, List, Optional, Tuple, Type, TypeVar, Union

from vellum import FunctionDefinition, PromptBlock, RichTextChildBlock, VellumVariable
from vellum.workflows.descriptors.base import BaseDescriptor
from vellum.workflows.nodes import InlinePromptNode
from vellum.workflows.types.core import JsonObject
from vellum.workflows.utils.functions import compile_function_definition
from vellum.workflows.utils.uuids import uuid4_from_hash
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.utils import raise_if_descriptor
from vellum_ee.workflows.display.nodes.vellum.utils import create_node_input
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.utils.vellum import infer_vellum_variable_type
from vellum_ee.workflows.display.vellum import NodeInput


def _contains_descriptors(obj):
    """Check if an object contains any descriptors or references that need special handling."""
    if isinstance(obj, BaseDescriptor):
        return True
    elif isinstance(obj, dict):
        return any(_contains_descriptors(v) for v in obj.values())
    elif isinstance(obj, (list, tuple)):
        return any(_contains_descriptors(item) for item in obj)
    elif hasattr(obj, "__dict__"):
        return any(_contains_descriptors(getattr(obj, field)) for field in obj.__dict__)
    return False


_InlinePromptNodeType = TypeVar("_InlinePromptNodeType", bound=InlinePromptNode)


class BaseInlinePromptNodeDisplay(BaseNodeDisplay[_InlinePromptNodeType], Generic[_InlinePromptNodeType]):
    __serializable_inputs__ = {
        InlinePromptNode.prompt_inputs,
    }
    __unserializable_attributes__ = {
        InlinePromptNode.settings,
        InlinePromptNode.expand_meta,
        InlinePromptNode.request_options,
    }

    def serialize(
        self, display_context: WorkflowDisplayContext, error_output_id: Optional[UUID] = None, **_kwargs
    ) -> JsonObject:
        node = self._node
        node_id = self.node_id

        node_inputs, prompt_inputs = self._generate_node_and_prompt_inputs(node_id, node, display_context)
        input_variable_id_by_name = {prompt_input.key: prompt_input.id for prompt_input in prompt_inputs}

        output_display = self.output_display[node.Outputs.text]
        array_display = self.output_display[node.Outputs.results]
        json_display = self.output_display[node.Outputs.json]
        node_blocks = raise_if_descriptor(node.blocks) or []
        function_definitions = raise_if_descriptor(node.functions)

        ml_model = str(raise_if_descriptor(node.ml_model))

        has_descriptors = _contains_descriptors(node_blocks)

        blocks: list = []
        if not has_descriptors:
            blocks = [
                self._generate_prompt_block(block, input_variable_id_by_name, [i])
                for i, block in enumerate(node_blocks)
                if not isinstance(block, BaseDescriptor)
            ]

        functions = (
            [self._generate_function_tools(function, i) for i, function in enumerate(function_definitions)]
            if isinstance(function_definitions, list)
            else []
        )
        blocks.extend(functions)

        serialized_node: JsonObject = {
            "id": str(node_id),
            "type": "PROMPT",
            "inputs": [node_input.dict() for node_input in node_inputs],
            "data": {
                "label": self.label,
                "output_id": str(output_display.id),
                "error_output_id": str(error_output_id) if error_output_id else None,
                "array_output_id": str(array_display.id),
                "source_handle_id": str(self.get_source_handle_id(display_context.port_displays)),
                "target_handle_id": str(self.get_target_handle_id()),
                "variant": "INLINE",
                "exec_config": {
                    "parameters": self._serialize_parameters(node.parameters, display_context),
                    "input_variables": [prompt_input.dict() for prompt_input in prompt_inputs],
                    "prompt_template_block_data": {
                        "version": 1,
                        "blocks": blocks,
                    },
                },
                "ml_model_name": ml_model,
            },
            "display_data": self.get_display_data().dict(),
            "base": self.get_base().dict(),
            "definition": self.get_definition().dict(),
            "outputs": [
                {"id": str(json_display.id), "name": "json", "type": "JSON", "value": None},
                {"id": str(output_display.id), "name": "text", "type": "STRING", "value": None},
                {"id": str(array_display.id), "name": "results", "type": "ARRAY", "value": None},
            ],
            "ports": self.serialize_ports(display_context),
        }
        attributes = self._serialize_attributes(display_context)
        if attributes:
            serialized_node["attributes"] = attributes
        return serialized_node

    def _generate_node_and_prompt_inputs(
        self,
        node_id: UUID,
        node: Type[InlinePromptNode],
        display_context: WorkflowDisplayContext,
    ) -> Tuple[List[NodeInput], List[VellumVariable]]:
        value = raise_if_descriptor(node.prompt_inputs)

        node_inputs: List[NodeInput] = []
        prompt_inputs: List[VellumVariable] = []

        if not value:
            return node_inputs, prompt_inputs

        for variable_name, variable_value in value.items():
            node_input = create_node_input(
                node_id=node_id,
                input_name=variable_name,
                value=variable_value,
                display_context=display_context,
                input_id=self.node_input_ids_by_name.get(f"{InlinePromptNode.prompt_inputs.name}.{variable_name}")
                or self.node_input_ids_by_name.get(variable_name),
            )
            vellum_variable_type = infer_vellum_variable_type(variable_value)
            node_inputs.append(node_input)
            prompt_inputs.append(VellumVariable(id=str(node_input.id), key=variable_name, type=vellum_variable_type))

        return node_inputs, prompt_inputs

    def _generate_function_tools(self, function: Union[FunctionDefinition, Callable], index: int) -> JsonObject:
        normalized_functions = (
            function if isinstance(function, FunctionDefinition) else compile_function_definition(function)
        )
        return {
            "id": str(uuid4_from_hash(f"{self.node_id}-FUNCTION_DEFINITION-{index}")),
            "block_type": "FUNCTION_DEFINITION",
            "properties": {
                "function_name": normalized_functions.name,
                "function_description": normalized_functions.description,
                "function_parameters": normalized_functions.parameters,
                "function_forced": normalized_functions.forced,
                "function_strict": normalized_functions.strict,
            },
        }

    def _generate_prompt_block(
        self,
        prompt_block: Union[PromptBlock, RichTextChildBlock],
        input_variable_id_by_name: Dict[str, str],
        path: List[int],
    ) -> JsonObject:
        block: JsonObject
        block_id = uuid4_from_hash(f"{self.node_id}-{prompt_block.block_type}-{'-'.join([str(i) for i in path])}")
        if prompt_block.block_type == "JINJA":
            block = {
                "block_type": "JINJA",
                "properties": {"template": prompt_block.template, "template_type": "STRING"},
            }

        elif prompt_block.block_type == "CHAT_MESSAGE":
            chat_properties: JsonObject = {
                "chat_role": prompt_block.chat_role,
                "chat_source": prompt_block.chat_source,
                "chat_message_unterminated": bool(prompt_block.chat_message_unterminated),
                "blocks": [
                    self._generate_prompt_block(block, input_variable_id_by_name, path + [i])
                    for i, block in enumerate(prompt_block.blocks)
                ],
            }

            block = {
                "block_type": "CHAT_MESSAGE",
                "properties": chat_properties,
            }

        elif prompt_block.block_type == "FUNCTION_DEFINITION":
            block = {
                "block_type": "FUNCTION_DEFINITION",
                "properties": {
                    "function_name": prompt_block.function_name,
                    "function_description": prompt_block.function_description,
                    "function_parameters": prompt_block.function_parameters,
                    "function_forced": prompt_block.function_forced,
                    "function_strict": prompt_block.function_strict,
                },
            }

        elif prompt_block.block_type == "VARIABLE":
            input_variable_id = input_variable_id_by_name.get(prompt_block.input_variable)
            if input_variable_id:
                block = {
                    "block_type": "VARIABLE",
                    "input_variable_id": input_variable_id,
                }
            else:
                # Even though this will likely fail in runtime, we want to allow serialization to succeed
                # in case the block is work in progress or the node is not yet part of the graph
                block = {
                    "block_type": "VARIABLE",
                    "input_variable_id": str(
                        uuid4_from_hash(f"{block_id}-input_variable-{prompt_block.input_variable}")
                    ),
                }

        elif prompt_block.block_type == "PLAIN_TEXT":
            block = {
                "block_type": "PLAIN_TEXT",
                "text": prompt_block.text,
            }

        elif prompt_block.block_type == "RICH_TEXT":
            block = {
                "block_type": "RICH_TEXT",
                "blocks": [
                    self._generate_prompt_block(child, input_variable_id_by_name, path + [i])
                    for i, child in enumerate(prompt_block.blocks)
                ],
            }
        else:
            raise NotImplementedError(f"Serialization for prompt block type {prompt_block.block_type} not implemented")

        block["id"] = str(block_id)
        if prompt_block.cache_config:
            block["cache_config"] = prompt_block.cache_config.dict()
        else:
            block["cache_config"] = None

        if prompt_block.state:
            block["state"] = prompt_block.state
        else:
            block["state"] = "ENABLED"

        return block

    def _serialize_parameters(self, parameters, display_context: "WorkflowDisplayContext") -> JsonObject:
        """Serialize parameters, returning empty object when nested descriptors are detected."""
        params = raise_if_descriptor(parameters)
        if not params:
            return {}

        if _contains_descriptors(params):
            return {}

        return params.dict()



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .deployment_release_tag_read import DeploymentReleaseTagRead
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class PaginatedDeploymentReleaseTagReadList(UniversalBaseModel):
    count: typing.Optional[int] = None
    next: typing.Optional[str] = None
    previous: typing.Optional[str] = None
    results: typing.Optional[typing.List[DeploymentReleaseTagRead]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from .workflow_execution_initiated_event import WorkflowExecutionInitiatedEvent
from .workflow_execution_streaming_event import WorkflowExecutionStreamingEvent
from .workflow_execution_fulfilled_event import WorkflowExecutionFulfilledEvent
from .workflow_execution_rejected_event import WorkflowExecutionRejectedEvent
from .workflow_execution_paused_event import WorkflowExecutionPausedEvent
from .workflow_execution_resumed_event import WorkflowExecutionResumedEvent
from .workflow_execution_snapshotted_event import WorkflowExecutionSnapshottedEvent

VellumWorkflowExecutionEvent = typing.Union[
    WorkflowExecutionInitiatedEvent,
    WorkflowExecutionStreamingEvent,
    WorkflowExecutionFulfilledEvent,
    WorkflowExecutionRejectedEvent,
    WorkflowExecutionPausedEvent,
    WorkflowExecutionResumedEvent,
    WorkflowExecutionSnapshottedEvent,
]



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .vellum_error_code_enum import VellumErrorCodeEnum
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import typing
import pydantic


class VellumError(UniversalBaseModel):
    message: str
    code: VellumErrorCodeEnum

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class CompilePromptMeta(UniversalBaseModel):
    """
    The subset of the metadata tracked by Vellum during Prompt Deployment compilation that the request opted into with `expand_meta`.
    """

    model_name: typing.Optional[str] = None
    deployment_release_tag: typing.Optional[str] = None
    prompt_version_id: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from ...core.request_options import RequestOptions
from ...core.http_response import HttpResponse
from ...types.paginated_test_suite_test_case_list import PaginatedTestSuiteTestCaseList
from ...core.jsonable_encoder import jsonable_encoder
from ...core.pydantic_utilities import parse_obj_as
from json.decoder import JSONDecodeError
from ...core.api_error import ApiError
from ...types.named_test_case_variable_value_request import NamedTestCaseVariableValueRequest
from ...types.test_suite_test_case import TestSuiteTestCase
from ...core.serialization import convert_and_respect_annotation_metadata
from ...types.test_suite_test_case_bulk_operation_request import TestSuiteTestCaseBulkOperationRequest
from ...types.test_suite_test_case_bulk_result import TestSuiteTestCaseBulkResult
import json
import contextlib
from ...core.client_wrapper import AsyncClientWrapper
from ...core.http_response import AsyncHttpResponse

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawTestSuitesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def list_test_suite_test_cases(
        self,
        id: str,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[PaginatedTestSuiteTestCaseList]:
        """
        List the Test Cases associated with a Test Suite

        Parameters
        ----------
        id : str
            Either the Test Suites' ID or its unique name

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[PaginatedTestSuiteTestCaseList]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/test-suites/{jsonable_encoder(id)}/test-cases",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "limit": limit,
                "offset": offset,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PaginatedTestSuiteTestCaseList,
                    parse_obj_as(
                        type_=PaginatedTestSuiteTestCaseList,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    def upsert_test_suite_test_case(
        self,
        id_: str,
        *,
        input_values: typing.Sequence[NamedTestCaseVariableValueRequest],
        evaluation_values: typing.Sequence[NamedTestCaseVariableValueRequest],
        id: typing.Optional[str] = OMIT,
        external_id: typing.Optional[str] = OMIT,
        label: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[TestSuiteTestCase]:
        """
        Upserts a new test case for a test suite, keying off of the optionally provided test case id.

        If an id is provided and has a match, the test case will be updated. If no id is provided or no match
        is found, a new test case will be appended to the end.

        Note that a full replacement of the test case is performed, so any fields not provided will be removed
        or overwritten with default values.

        Parameters
        ----------
        id_ : str
            Either the Test Suites' ID or its unique name

        input_values : typing.Sequence[NamedTestCaseVariableValueRequest]
            Values for each of the Test Case's input variables

        evaluation_values : typing.Sequence[NamedTestCaseVariableValueRequest]
            Values for each of the Test Case's evaluation variables

        id : typing.Optional[str]
            The Vellum-generated ID of an existing Test Case whose data you'd like to replace. If specified and no Test Case exists with this ID, a 404 will be returned.

        external_id : typing.Optional[str]
            An ID external to Vellum that uniquely identifies the Test Case that you'd like to create/update. If there's a match on a Test Case that was previously created with the same external_id, it will be updated. Otherwise, a new Test Case will be created with this value as its external_id. If no external_id is specified, then a new Test Case will always be created.

        label : typing.Optional[str]
            A human-readable label used to convey the intention of this Test Case

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[TestSuiteTestCase]

        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/test-suites/{jsonable_encoder(id_)}/test-cases",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "id": id,
                "external_id": external_id,
                "label": label,
                "input_values": convert_and_respect_annotation_metadata(
                    object_=input_values,
                    annotation=typing.Sequence[NamedTestCaseVariableValueRequest],
                    direction="write",
                ),
                "evaluation_values": convert_and_respect_annotation_metadata(
                    object_=evaluation_values,
                    annotation=typing.Sequence[NamedTestCaseVariableValueRequest],
                    direction="write",
                ),
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TestSuiteTestCase,
                    parse_obj_as(
                        type_=TestSuiteTestCase,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    @contextlib.contextmanager
    def test_suite_test_cases_bulk(
        self,
        id: str,
        *,
        request: typing.Sequence[TestSuiteTestCaseBulkOperationRequest],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.Iterator[HttpResponse[typing.Iterator[typing.List[TestSuiteTestCaseBulkResult]]]]:
        """
        Created, replace, and delete Test Cases within the specified Test Suite in bulk

        Parameters
        ----------
        id : str
            Either the Test Suites' ID or its unique name

        request : typing.Sequence[TestSuiteTestCaseBulkOperationRequest]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Yields
        ------
        typing.Iterator[HttpResponse[typing.Iterator[typing.List[TestSuiteTestCaseBulkResult]]]]

        """
        with self._client_wrapper.httpx_client.stream(
            f"v1/test-suites/{jsonable_encoder(id)}/test-cases-bulk",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json=convert_and_respect_annotation_metadata(
                object_=request, annotation=typing.Sequence[TestSuiteTestCaseBulkOperationRequest], direction="write"
            ),
            request_options=request_options,
            omit=OMIT,
        ) as _response:

            def stream() -> HttpResponse[typing.Iterator[typing.List[TestSuiteTestCaseBulkResult]]]:
                try:
                    if 200 <= _response.status_code < 300:

                        def _iter():
                            for _text in _response.iter_lines():
                                try:
                                    if len(_text) == 0:
                                        continue
                                    yield typing.cast(
                                        typing.List[TestSuiteTestCaseBulkResult],
                                        parse_obj_as(
                                            type_=typing.List[TestSuiteTestCaseBulkResult],  # type: ignore
                                            object_=json.loads(_text),
                                        ),
                                    )
                                except Exception:
                                    pass
                            return

                        return HttpResponse(response=_response, data=_iter())
                    _response.read()
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(status_code=_response.status_code, body=_response.text)
                raise ApiError(status_code=_response.status_code, body=_response_json)

            yield stream()

    def delete_test_suite_test_case(
        self, id: str, test_case_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[None]:
        """
        Deletes an existing test case for a test suite, keying off of the test case id.

        Parameters
        ----------
        id : str
            Either the Test Suites' ID or its unique name

        test_case_id : str
            An id identifying the test case that you'd like to delete

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/test-suites/{jsonable_encoder(id)}/test-cases/{jsonable_encoder(test_case_id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)


class AsyncRawTestSuitesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def list_test_suite_test_cases(
        self,
        id: str,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[PaginatedTestSuiteTestCaseList]:
        """
        List the Test Cases associated with a Test Suite

        Parameters
        ----------
        id : str
            Either the Test Suites' ID or its unique name

        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[PaginatedTestSuiteTestCaseList]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/test-suites/{jsonable_encoder(id)}/test-cases",
            base_url=self._client_wrapper.get_environment().default,
            method="GET",
            params={
                "limit": limit,
                "offset": offset,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    PaginatedTestSuiteTestCaseList,
                    parse_obj_as(
                        type_=PaginatedTestSuiteTestCaseList,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    async def upsert_test_suite_test_case(
        self,
        id_: str,
        *,
        input_values: typing.Sequence[NamedTestCaseVariableValueRequest],
        evaluation_values: typing.Sequence[NamedTestCaseVariableValueRequest],
        id: typing.Optional[str] = OMIT,
        external_id: typing.Optional[str] = OMIT,
        label: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[TestSuiteTestCase]:
        """
        Upserts a new test case for a test suite, keying off of the optionally provided test case id.

        If an id is provided and has a match, the test case will be updated. If no id is provided or no match
        is found, a new test case will be appended to the end.

        Note that a full replacement of the test case is performed, so any fields not provided will be removed
        or overwritten with default values.

        Parameters
        ----------
        id_ : str
            Either the Test Suites' ID or its unique name

        input_values : typing.Sequence[NamedTestCaseVariableValueRequest]
            Values for each of the Test Case's input variables

        evaluation_values : typing.Sequence[NamedTestCaseVariableValueRequest]
            Values for each of the Test Case's evaluation variables

        id : typing.Optional[str]
            The Vellum-generated ID of an existing Test Case whose data you'd like to replace. If specified and no Test Case exists with this ID, a 404 will be returned.

        external_id : typing.Optional[str]
            An ID external to Vellum that uniquely identifies the Test Case that you'd like to create/update. If there's a match on a Test Case that was previously created with the same external_id, it will be updated. Otherwise, a new Test Case will be created with this value as its external_id. If no external_id is specified, then a new Test Case will always be created.

        label : typing.Optional[str]
            A human-readable label used to convey the intention of this Test Case

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[TestSuiteTestCase]

        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/test-suites/{jsonable_encoder(id_)}/test-cases",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json={
                "id": id,
                "external_id": external_id,
                "label": label,
                "input_values": convert_and_respect_annotation_metadata(
                    object_=input_values,
                    annotation=typing.Sequence[NamedTestCaseVariableValueRequest],
                    direction="write",
                ),
                "evaluation_values": convert_and_respect_annotation_metadata(
                    object_=evaluation_values,
                    annotation=typing.Sequence[NamedTestCaseVariableValueRequest],
                    direction="write",
                ),
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    TestSuiteTestCase,
                    parse_obj_as(
                        type_=TestSuiteTestCase,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)

    @contextlib.asynccontextmanager
    async def test_suite_test_cases_bulk(
        self,
        id: str,
        *,
        request: typing.Sequence[TestSuiteTestCaseBulkOperationRequest],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[typing.List[TestSuiteTestCaseBulkResult]]]]:
        """
        Created, replace, and delete Test Cases within the specified Test Suite in bulk

        Parameters
        ----------
        id : str
            Either the Test Suites' ID or its unique name

        request : typing.Sequence[TestSuiteTestCaseBulkOperationRequest]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Yields
        ------
        typing.AsyncIterator[AsyncHttpResponse[typing.AsyncIterator[typing.List[TestSuiteTestCaseBulkResult]]]]

        """
        async with self._client_wrapper.httpx_client.stream(
            f"v1/test-suites/{jsonable_encoder(id)}/test-cases-bulk",
            base_url=self._client_wrapper.get_environment().default,
            method="POST",
            json=convert_and_respect_annotation_metadata(
                object_=request, annotation=typing.Sequence[TestSuiteTestCaseBulkOperationRequest], direction="write"
            ),
            request_options=request_options,
            omit=OMIT,
        ) as _response:

            async def stream() -> AsyncHttpResponse[typing.AsyncIterator[typing.List[TestSuiteTestCaseBulkResult]]]:
                try:
                    if 200 <= _response.status_code < 300:

                        async def _iter():
                            async for _text in _response.aiter_lines():
                                try:
                                    if len(_text) == 0:
                                        continue
                                    yield typing.cast(
                                        typing.List[TestSuiteTestCaseBulkResult],
                                        parse_obj_as(
                                            type_=typing.List[TestSuiteTestCaseBulkResult],  # type: ignore
                                            object_=json.loads(_text),
                                        ),
                                    )
                                except Exception:
                                    pass
                            return

                        return AsyncHttpResponse(response=_response, data=_iter())
                    await _response.aread()
                    _response_json = _response.json()
                except JSONDecodeError:
                    raise ApiError(status_code=_response.status_code, body=_response.text)
                raise ApiError(status_code=_response.status_code, body=_response_json)

            yield await stream()

    async def delete_test_suite_test_case(
        self, id: str, test_case_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Deletes an existing test case for a test suite, keying off of the test case id.

        Parameters
        ----------
        id : str
            Either the Test Suites' ID or its unique name

        test_case_id : str
            An id identifying the test case that you'd like to delete

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/test-suites/{jsonable_encoder(id)}/test-cases/{jsonable_encoder(test_case_id)}",
            base_url=self._client_wrapper.get_environment().default,
            method="DELETE",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, body=_response.text)
        raise ApiError(status_code=_response.status_code, body=_response_json)



from deepdiff import DeepDiff

from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.types.core import MergeBehavior


class Inputs(BaseInputs):
    input: str


def test_serialize_node__basic(serialize_node):
    class BasicGenericNode(BaseNode):
        pass

    serialized_node = serialize_node(BasicGenericNode)
    assert not DeepDiff(
        {
            "id": "8d7cbfe4-72ca-4367-a401-8d28723d2f00",
            "label": "test_serialize_node__basic.<locals>.BasicGenericNode",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "BasicGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_trigger_serialization",
                ],
            },
            "trigger": {"id": "b95cca96-b570-42ac-ace8-51ca0f627881", "merge_behavior": "AWAIT_ATTRIBUTES"},
            "ports": [
                {
                    "id": "8bec8d0c-113f-4110-afcb-4a6e566e7236",
                    "name": "default",
                    "type": "DEFAULT",
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__await_any(serialize_node):
    class AwaitAnyGenericNode(BaseNode):
        class Trigger(BaseNode.Trigger):
            merge_behavior = MergeBehavior.AWAIT_ANY

    serialized_node = serialize_node(AwaitAnyGenericNode)
    assert not DeepDiff(
        {
            "id": "42e17f0e-8496-415f-9c72-f85250ba6f0b",
            "label": "test_serialize_node__await_any.<locals>.AwaitAnyGenericNode",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "AwaitAnyGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_trigger_serialization",
                ],
            },
            "trigger": {"id": "c0db17e7-6766-4062-aaee-7404580d76e4", "merge_behavior": "AWAIT_ANY"},
            "ports": [
                {
                    "id": "d9a84db7-8bd6-4a15-9e3c-c2e898c26d16",
                    "name": "default",
                    "type": "DEFAULT",
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )


def test_serialize_node__await_all(serialize_node):
    class AwaitAllGenericNode(BaseNode):
        class Trigger(BaseNode.Trigger):
            merge_behavior = MergeBehavior.AWAIT_ALL

    serialized_node = serialize_node(AwaitAllGenericNode)
    assert not DeepDiff(
        {
            "id": "b3e1145a-5f41-456b-9382-6d0a1e828c2f",
            "label": "test_serialize_node__await_all.<locals>.AwaitAllGenericNode",
            "type": "GENERIC",
            "display_data": {"position": {"x": 0.0, "y": 0.0}},
            "base": {"name": "BaseNode", "module": ["vellum", "workflows", "nodes", "bases", "base"]},
            "definition": {
                "name": "AwaitAllGenericNode",
                "module": [
                    "vellum_ee",
                    "workflows",
                    "display",
                    "tests",
                    "workflow_serialization",
                    "generic_nodes",
                    "test_trigger_serialization",
                ],
            },
            "trigger": {"id": "1b22935e-0e79-485a-b274-a2f316c0983c", "merge_behavior": "AWAIT_ALL"},
            "ports": [
                {
                    "id": "fa73da35-0bf9-4f02-bf5b-0b0d1a6f1494",
                    "name": "default",
                    "type": "DEFAULT",
                }
            ],
            "adornments": None,
            "attributes": [],
            "outputs": [],
        },
        serialized_node,
        ignore_order=True,
    )



# This file was auto-generated by Fern from our API Definition.

import typing
from .string_input import StringInput
from .json_input import JsonInput
from .chat_history_input import ChatHistoryInput
from .number_input import NumberInput
from .search_results_input import SearchResultsInput
from .error_input import ErrorInput
from .array_input import ArrayInput
from .function_call_input import FunctionCallInput
from .code_executor_secret_input import CodeExecutorSecretInput

CodeExecutorInput = typing.Union[
    StringInput,
    JsonInput,
    ChatHistoryInput,
    NumberInput,
    SearchResultsInput,
    ErrorInput,
    ArrayInput,
    FunctionCallInput,
    CodeExecutorSecretInput,
]



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
import datetime as dt
from .new_member_join_behavior_enum import NewMemberJoinBehaviorEnum
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class OrganizationRead(UniversalBaseModel):
    id: str
    name: str
    created: typing.Optional[dt.datetime] = None
    allow_staff_access: typing.Optional[bool] = None
    new_member_join_behavior: NewMemberJoinBehaviorEnum
    limit_config: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import inspect
from uuid import UUID
from typing import ClassVar, Generic, Optional, TypeVar

from vellum.workflows.nodes.displayable.code_execution_node import CodeExecutionNode
from vellum.workflows.nodes.displayable.code_execution_node.utils import read_file_from_path
from vellum.workflows.types.core import JsonObject
from vellum.workflows.utils.vellum_variables import primitive_type_to_vellum_variable_type
from vellum_ee.workflows.display.exceptions import NodeValidationError
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.utils import raise_if_descriptor
from vellum_ee.workflows.display.nodes.vellum.utils import create_node_input
from vellum_ee.workflows.display.types import WorkflowDisplayContext

_CodeExecutionNodeType = TypeVar("_CodeExecutionNodeType", bound=CodeExecutionNode)


class BaseCodeExecutionNodeDisplay(BaseNodeDisplay[_CodeExecutionNodeType], Generic[_CodeExecutionNodeType]):
    output_id: ClassVar[Optional[UUID]] = None
    log_output_id: ClassVar[Optional[UUID]] = None

    __serializable_inputs__ = {
        CodeExecutionNode.code,
        CodeExecutionNode.code_inputs,
    }

    def serialize(
        self, display_context: WorkflowDisplayContext, error_output_id: Optional[UUID] = None, **_kwargs
    ) -> JsonObject:
        node = self._node
        node_id = self.node_id
        raw_code = raise_if_descriptor(node.code)
        filepath = raise_if_descriptor(node.filepath)

        code_value: Optional[str]
        if raw_code:
            code_value = raw_code
        elif filepath:
            node_file_path = inspect.getfile(node)
            file_code = read_file_from_path(
                node_filepath=node_file_path,
                script_filepath=filepath,
            )
            if not file_code:
                display_context.add_error(NodeValidationError(f"Filepath '{filepath}' does not exist", node.__name__))
            code_value = file_code
        else:
            code_value = ""

        code_inputs = raise_if_descriptor(node.code_inputs)

        inputs = [
            create_node_input(
                node_id=node_id,
                input_name=variable_name,
                value=variable_value,
                display_context=display_context,
                input_id=self.node_input_ids_by_name.get(f"{CodeExecutionNode.code_inputs.name}.{variable_name}")
                or self.node_input_ids_by_name.get(variable_name),
            )
            for variable_name, variable_value in code_inputs.items()
        ]

        code_input_id = self.node_input_ids_by_name.get(CodeExecutionNode.code.name)
        code_node_input = create_node_input(
            node_id=node_id,
            input_name="code",
            value=code_value,
            display_context=display_context,
            input_id=code_input_id,
        )

        runtime_input_id = self.node_input_ids_by_name.get(CodeExecutionNode.runtime.name)
        runtime_node_input = create_node_input(
            node_id=node_id,
            input_name="runtime",
            value=node.runtime,
            display_context=display_context,
            input_id=runtime_input_id,
        )
        inputs.extend([code_node_input, runtime_node_input])

        packages = raise_if_descriptor(node.packages)

        output_display = self.output_display[node.Outputs.result]
        log_output_display = self.output_display[node.Outputs.log]

        output_type = primitive_type_to_vellum_variable_type(node.get_output_type())

        return {
            "id": str(node_id),
            "type": "CODE_EXECUTION",
            "inputs": [input.dict() for input in inputs],
            "data": {
                "label": self.label,
                "error_output_id": str(error_output_id) if error_output_id else None,
                "source_handle_id": str(self.get_source_handle_id(display_context.port_displays)),
                "target_handle_id": str(self.get_target_handle_id()),
                "code_input_id": code_node_input.id,
                "runtime_input_id": runtime_node_input.id,
                "output_type": output_type,
                "packages": [package.dict() for package in packages] if packages is not None else [],
                "output_id": str(self.output_id) if self.output_id else str(output_display.id),
                "log_output_id": str(self.log_output_id) if self.log_output_id else str(log_output_display.id),
            },
            "display_data": self.get_display_data().dict(),
            "base": self.get_base().dict(),
            "definition": self.get_definition().dict(),
            "ports": self.serialize_ports(display_context),
        }



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
import typing
from .prompt_block_state import PromptBlockState
from .ephemeral_prompt_cache_config import EphemeralPromptCacheConfig
import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class FunctionDefinition(UniversalBaseModel):
    """
    The definition of a Function (aka "Tool Call") that a Prompt/Model has access to.
    """

    state: typing.Optional[PromptBlockState] = None
    cache_config: typing.Optional[EphemeralPromptCacheConfig] = None
    name: typing.Optional[str] = pydantic.Field(default=None)
    """
    The name identifying the function.
    """

    description: typing.Optional[str] = pydantic.Field(default=None)
    """
    A description to help guide the model when to invoke this function.
    """

    parameters: typing.Optional[typing.Dict[str, typing.Optional[typing.Any]]] = pydantic.Field(default=None)
    """
    An OpenAPI specification of parameters that are supported by this function.
    """

    forced: typing.Optional[bool] = pydantic.Field(default=None)
    """
    Set this option to true to force the model to return a function call of this function.
    """

    strict: typing.Optional[bool] = pydantic.Field(default=None)
    """
    Set this option to use strict schema decoding when available.
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

import typing
from ...core.client_wrapper import SyncClientWrapper
from .raw_client import RawDocumentIndexesClient
from .types.document_indexes_list_request_status import DocumentIndexesListRequestStatus
from ...core.request_options import RequestOptions
from ...types.paginated_document_index_read_list import PaginatedDocumentIndexReadList
from ...types.document_index_indexing_config_request import DocumentIndexIndexingConfigRequest
from ...types.entity_status import EntityStatus
from ...types.document_index_read import DocumentIndexRead
from ...core.client_wrapper import AsyncClientWrapper
from .raw_client import AsyncRawDocumentIndexesClient

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class DocumentIndexesClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawDocumentIndexesClient(client_wrapper=client_wrapper)

    @property
    def _client_wrapper(self) -> SyncClientWrapper:
        return self._raw_client._client_wrapper

    @property
    def with_raw_response(self) -> RawDocumentIndexesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawDocumentIndexesClient
        """
        return self._raw_client

    def list(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        status: typing.Optional[DocumentIndexesListRequestStatus] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedDocumentIndexReadList:
        """
        Used to retrieve a list of Document Indexes.

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        search : typing.Optional[str]
            Search for document indices by name or label

        status : typing.Optional[DocumentIndexesListRequestStatus]
            Filter down to only document indices that have a status matching the status specified

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedDocumentIndexReadList


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.document_indexes.list()
        """
        response = self._raw_client.list(
            limit=limit, offset=offset, ordering=ordering, search=search, status=status, request_options=request_options
        )
        return response.data

    def create(
        self,
        *,
        label: str,
        name: str,
        indexing_config: DocumentIndexIndexingConfigRequest,
        status: typing.Optional[EntityStatus] = OMIT,
        copy_documents_from_index_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DocumentIndexRead:
        """
        Creates a new document index.

        Parameters
        ----------
        label : str
            A human-readable label for the document index

        name : str
            A name that uniquely identifies this index within its workspace

        indexing_config : DocumentIndexIndexingConfigRequest

        status : typing.Optional[EntityStatus]
            The current status of the document index

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        copy_documents_from_index_id : typing.Optional[str]
            Optionally specify the id of a document index from which you'd like to copy and re-index its documents into this newly created index

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentIndexRead


        Examples
        --------
        from vellum import (
            DocumentIndexIndexingConfigRequest,
            OpenAiVectorizerConfigRequest,
            OpenAiVectorizerTextEmbedding3SmallRequest,
            Vellum,
        )

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.document_indexes.create(
            label="x",
            name="x",
            indexing_config=DocumentIndexIndexingConfigRequest(
                vectorizer=OpenAiVectorizerTextEmbedding3SmallRequest(
                    config=OpenAiVectorizerConfigRequest(),
                ),
            ),
        )
        """
        response = self._raw_client.create(
            label=label,
            name=name,
            indexing_config=indexing_config,
            status=status,
            copy_documents_from_index_id=copy_documents_from_index_id,
            request_options=request_options,
        )
        return response.data

    def retrieve(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> DocumentIndexRead:
        """
        Used to retrieve a Document Index given its ID or name.

        Parameters
        ----------
        id : str
            Either the Document Index's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentIndexRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.document_indexes.retrieve(
            id="id",
        )
        """
        response = self._raw_client.retrieve(id, request_options=request_options)
        return response.data

    def update(
        self,
        id: str,
        *,
        label: str,
        status: typing.Optional[EntityStatus] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DocumentIndexRead:
        """
        Used to fully update a Document Index given its ID or name.

        Parameters
        ----------
        id : str
            Either the Document Index's ID or its unique name

        label : str
            A human-readable label for the document index

        status : typing.Optional[EntityStatus]
            The current status of the document index

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentIndexRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.document_indexes.update(
            id="id",
            label="x",
        )
        """
        response = self._raw_client.update(id, label=label, status=status, request_options=request_options)
        return response.data

    def destroy(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Used to delete a Document Index given its ID or name.

        Parameters
        ----------
        id : str
            Either the Document Index's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.document_indexes.destroy(
            id="id",
        )
        """
        response = self._raw_client.destroy(id, request_options=request_options)
        return response.data

    def partial_update(
        self,
        id: str,
        *,
        label: typing.Optional[str] = OMIT,
        status: typing.Optional[EntityStatus] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DocumentIndexRead:
        """
        Used to partial update a Document Index given its ID or name.

        Parameters
        ----------
        id : str
            Either the Document Index's ID or its unique name

        label : typing.Optional[str]
            A human-readable label for the document index

        status : typing.Optional[EntityStatus]
            The current status of the document index

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentIndexRead


        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.document_indexes.partial_update(
            id="id",
        )
        """
        response = self._raw_client.partial_update(id, label=label, status=status, request_options=request_options)
        return response.data

    def add_document(
        self, document_id: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Adds a previously uploaded Document to the specified Document Index.

        Parameters
        ----------
        document_id : str
            Either the Vellum-generated ID or the originally supplied external_id that uniquely identifies the Document you'd like to add.

        id : str
            Either the Vellum-generated ID or the originally specified name that uniquely identifies the Document Index to which you'd like to add the Document.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.document_indexes.add_document(
            document_id="document_id",
            id="id",
        )
        """
        response = self._raw_client.add_document(document_id, id, request_options=request_options)
        return response.data

    def remove_document(
        self, document_id: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Removes a Document from a Document Index without deleting the Document itself.

        Parameters
        ----------
        document_id : str
            Either the Vellum-generated ID or the originally supplied external_id that uniquely identifies the Document you'd like to remove.

        id : str
            Either the Vellum-generated ID or the originally specified name that uniquely identifies the Document Index from which you'd like to remove a Document.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from vellum import Vellum

        client = Vellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )
        client.document_indexes.remove_document(
            document_id="document_id",
            id="id",
        )
        """
        response = self._raw_client.remove_document(document_id, id, request_options=request_options)
        return response.data


class AsyncDocumentIndexesClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawDocumentIndexesClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawDocumentIndexesClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawDocumentIndexesClient
        """
        return self._raw_client

    async def list(
        self,
        *,
        limit: typing.Optional[int] = None,
        offset: typing.Optional[int] = None,
        ordering: typing.Optional[str] = None,
        search: typing.Optional[str] = None,
        status: typing.Optional[DocumentIndexesListRequestStatus] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PaginatedDocumentIndexReadList:
        """
        Used to retrieve a list of Document Indexes.

        Parameters
        ----------
        limit : typing.Optional[int]
            Number of results to return per page.

        offset : typing.Optional[int]
            The initial index from which to return the results.

        ordering : typing.Optional[str]
            Which field to use when ordering the results.

        search : typing.Optional[str]
            Search for document indices by name or label

        status : typing.Optional[DocumentIndexesListRequestStatus]
            Filter down to only document indices that have a status matching the status specified

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PaginatedDocumentIndexReadList


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document_indexes.list()


        asyncio.run(main())
        """
        response = await self._raw_client.list(
            limit=limit, offset=offset, ordering=ordering, search=search, status=status, request_options=request_options
        )
        return response.data

    async def create(
        self,
        *,
        label: str,
        name: str,
        indexing_config: DocumentIndexIndexingConfigRequest,
        status: typing.Optional[EntityStatus] = OMIT,
        copy_documents_from_index_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DocumentIndexRead:
        """
        Creates a new document index.

        Parameters
        ----------
        label : str
            A human-readable label for the document index

        name : str
            A name that uniquely identifies this index within its workspace

        indexing_config : DocumentIndexIndexingConfigRequest

        status : typing.Optional[EntityStatus]
            The current status of the document index

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        copy_documents_from_index_id : typing.Optional[str]
            Optionally specify the id of a document index from which you'd like to copy and re-index its documents into this newly created index

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentIndexRead


        Examples
        --------
        import asyncio

        from vellum import (
            AsyncVellum,
            DocumentIndexIndexingConfigRequest,
            OpenAiVectorizerConfigRequest,
            OpenAiVectorizerTextEmbedding3SmallRequest,
        )

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document_indexes.create(
                label="x",
                name="x",
                indexing_config=DocumentIndexIndexingConfigRequest(
                    vectorizer=OpenAiVectorizerTextEmbedding3SmallRequest(
                        config=OpenAiVectorizerConfigRequest(),
                    ),
                ),
            )


        asyncio.run(main())
        """
        response = await self._raw_client.create(
            label=label,
            name=name,
            indexing_config=indexing_config,
            status=status,
            copy_documents_from_index_id=copy_documents_from_index_id,
            request_options=request_options,
        )
        return response.data

    async def retrieve(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> DocumentIndexRead:
        """
        Used to retrieve a Document Index given its ID or name.

        Parameters
        ----------
        id : str
            Either the Document Index's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentIndexRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document_indexes.retrieve(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.retrieve(id, request_options=request_options)
        return response.data

    async def update(
        self,
        id: str,
        *,
        label: str,
        status: typing.Optional[EntityStatus] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DocumentIndexRead:
        """
        Used to fully update a Document Index given its ID or name.

        Parameters
        ----------
        id : str
            Either the Document Index's ID or its unique name

        label : str
            A human-readable label for the document index

        status : typing.Optional[EntityStatus]
            The current status of the document index

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentIndexRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document_indexes.update(
                id="id",
                label="x",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.update(id, label=label, status=status, request_options=request_options)
        return response.data

    async def destroy(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Used to delete a Document Index given its ID or name.

        Parameters
        ----------
        id : str
            Either the Document Index's ID or its unique name

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document_indexes.destroy(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.destroy(id, request_options=request_options)
        return response.data

    async def partial_update(
        self,
        id: str,
        *,
        label: typing.Optional[str] = OMIT,
        status: typing.Optional[EntityStatus] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> DocumentIndexRead:
        """
        Used to partial update a Document Index given its ID or name.

        Parameters
        ----------
        id : str
            Either the Document Index's ID or its unique name

        label : typing.Optional[str]
            A human-readable label for the document index

        status : typing.Optional[EntityStatus]
            The current status of the document index

            * `ACTIVE` - Active
            * `ARCHIVED` - Archived

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        DocumentIndexRead


        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document_indexes.partial_update(
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.partial_update(
            id, label=label, status=status, request_options=request_options
        )
        return response.data

    async def add_document(
        self, document_id: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Adds a previously uploaded Document to the specified Document Index.

        Parameters
        ----------
        document_id : str
            Either the Vellum-generated ID or the originally supplied external_id that uniquely identifies the Document you'd like to add.

        id : str
            Either the Vellum-generated ID or the originally specified name that uniquely identifies the Document Index to which you'd like to add the Document.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document_indexes.add_document(
                document_id="document_id",
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.add_document(document_id, id, request_options=request_options)
        return response.data

    async def remove_document(
        self, document_id: str, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Removes a Document from a Document Index without deleting the Document itself.

        Parameters
        ----------
        document_id : str
            Either the Vellum-generated ID or the originally supplied external_id that uniquely identifies the Document you'd like to remove.

        id : str
            Either the Vellum-generated ID or the originally specified name that uniquely identifies the Document Index from which you'd like to remove a Document.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        import asyncio

        from vellum import AsyncVellum

        client = AsyncVellum(
            api_version="YOUR_API_VERSION",
            api_key="YOUR_API_KEY",
        )


        async def main() -> None:
            await client.document_indexes.remove_document(
                document_id="document_id",
                id="id",
            )


        asyncio.run(main())
        """
        response = await self._raw_client.remove_document(document_id, id, request_options=request_options)
        return response.data



from vellum.workflows.nodes.displayable import TemplatingNode as BaseTemplatingNode
from vellum.workflows.state import BaseState

from ..inputs import Inputs


class TemplatingNode(BaseTemplatingNode[BaseState, str]):
    template = """{{ example_var_1 }}"""
    inputs = {
        "example_var_1": Inputs.input_value,
    }



from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs import BaseInputs
from vellum.workflows.nodes.displayable.inline_prompt_node.node import InlinePromptNode
from vellum.workflows.nodes.displayable.tool_calling_node.node import ToolCallingNode
from vellum.workflows.state.base import BaseState
from vellum.workflows.types.definition import AuthorizationType, EnvironmentVariableReference, MCPServer
from vellum_ee.workflows.display.workflows.get_vellum_workflow_display_class import get_workflow_display


def test_serialize_node__prompt_inputs__constant_value():
    # GIVEN a prompt node with constant value inputs
    class MyPromptNode(ToolCallingNode):
        prompt_inputs = {"foo": "bar"}

    # AND a workflow with the prompt node
    class Workflow(BaseWorkflow):
        graph = MyPromptNode

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should properly serialize the inputs
    my_prompt_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["id"] == str(MyPromptNode.__id__)
    )

    prompt_inputs_attribute = next(
        attribute for attribute in my_prompt_node["attributes"] if attribute["name"] == "prompt_inputs"
    )

    assert prompt_inputs_attribute == {
        "id": "3d9a4d2e-c9bd-4417-8a0c-52f15efdbe30",
        "name": "prompt_inputs",
        "value": {"type": "CONSTANT_VALUE", "value": {"type": "JSON", "value": {"foo": "bar"}}},
    }


def test_serialize_node__prompt_inputs__input_reference():
    # GIVEN a state definition
    class MyInput(BaseInputs):
        foo: str

    # AND a prompt node with inputs
    class MyPromptNode(InlinePromptNode):
        prompt_inputs = {"foo": MyInput.foo}

    # AND a workflow with the prompt node
    class Workflow(BaseWorkflow[MyInput, BaseState]):
        graph = MyPromptNode

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should skip the state reference input rule
    my_prompt_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["id"] == str(MyPromptNode.__id__)
    )

    prompt_inputs_attribute = next(
        attribute for attribute in my_prompt_node["attributes"] if attribute["name"] == "prompt_inputs"
    )

    assert prompt_inputs_attribute == {
        "id": "6cde4776-7f4a-411c-95a8-69c8b3a64b42",
        "name": "prompt_inputs",
        "value": {
            "type": "DICTIONARY_REFERENCE",
            "entries": [
                {
                    "id": "ab7902ef-de14-4edc-835c-366d3ef6a70e",
                    "key": "foo",
                    "value": {"type": "WORKFLOW_INPUT", "input_variable_id": "e3657390-fd3c-4fea-8cdd-fc5ea79f3278"},
                }
            ],
        },
    }


def test_serialize_node__prompt_inputs__mixed_values():
    # GIVEN a prompt node with mixed values
    class MyInput(BaseInputs):
        foo: str

    # AND a prompt node with mixed values
    class MyPromptNode(InlinePromptNode):
        prompt_inputs = {"foo": "bar", "baz": MyInput.foo}

    # AND a workflow with the prompt node
    class Workflow(BaseWorkflow[MyInput, BaseState]):
        graph = MyPromptNode

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should properly serialize the inputs
    my_prompt_node = next(
        node for node in serialized_workflow["workflow_raw_data"]["nodes"] if node["id"] == str(MyPromptNode.__id__)
    )

    prompt_inputs_attribute = next(
        attribute for attribute in my_prompt_node["attributes"] if attribute["name"] == "prompt_inputs"
    )

    assert prompt_inputs_attribute == {
        "id": "c4ca6e3d-0f71-4802-a618-1e87880cb7cf",
        "name": "prompt_inputs",
        "value": {
            "type": "DICTIONARY_REFERENCE",
            "entries": [
                {
                    "id": "0fc7e25e-075c-4849-b89b-9729d1aeada1",
                    "key": "foo",
                    "value": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "bar"}},
                },
                {
                    "id": "bba42c89-fa7b-4cb7-bc16-0d21ce060a4b",
                    "key": "baz",
                    "value": {"type": "WORKFLOW_INPUT", "input_variable_id": "8d57cf1d-147c-427b-9a5e-e5f6ab76e2eb"},
                },
            ],
        },
    }


def test_serialize_node__tool_calling_node__mcp_server_api_key():
    # GIVEN a tool calling node with an mcp server
    class MyToolCallingNode(ToolCallingNode):
        functions = [
            MCPServer(
                name="my-mcp-server",
                url="https://my-mcp-server.com",
                authorization_type=AuthorizationType.API_KEY,
                api_key_header_key="my-api-key-header-key",
                api_key_header_value=EnvironmentVariableReference(name="my-api-key-header-value"),
            )
        ]

    # AND a workflow with the tool calling node
    class Workflow(BaseWorkflow):
        graph = MyToolCallingNode

    # WHEN the workflow is serialized
    workflow_display = get_workflow_display(workflow_class=Workflow)
    serialized_workflow: dict = workflow_display.serialize()

    # THEN the node should properly serialize the mcp server
    my_tool_calling_node = next(
        node
        for node in serialized_workflow["workflow_raw_data"]["nodes"]
        if node["id"] == str(MyToolCallingNode.__id__)
    )

    functions_attribute = next(
        attribute for attribute in my_tool_calling_node["attributes"] if attribute["name"] == "functions"
    )

    assert functions_attribute == {
        "id": "6c0f7d4f-3c8a-4201-b588-8398d3c97480",
        "name": "functions",
        "value": {
            "type": "CONSTANT_VALUE",
            "value": {
                "type": "JSON",
                "value": [
                    {
                        "type": "MCP_SERVER",
                        "name": "my-mcp-server",
                        "url": "https://my-mcp-server.com",
                        "authorization_type": "API_KEY",
                        "bearer_token_value": None,
                        "api_key_header_key": "my-api-key-header-key",
                        "api_key_header_value": "my-api-key-header-value",
                    }
                ],
            },
        },
    }



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .api_request_parent_context import ApiRequestParentContext
from .external_parent_context import ExternalParentContext
from .node_parent_context import NodeParentContext
from .prompt_deployment_parent_context import PromptDeploymentParentContext
from .span_link import SpanLink
from .workflow_deployment_parent_context import WorkflowDeploymentParentContext
from .workflow_parent_context import WorkflowParentContext
from .workflow_sandbox_parent_context import WorkflowSandboxParentContext
import typing
from .vellum_node_execution_event import VellumNodeExecutionEvent
from .node_execution_span_attributes import NodeExecutionSpanAttributes
from .workflow_execution_usage_calculation_fulfilled_body import WorkflowExecutionUsageCalculationFulfilledBody
import datetime as dt
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class NodeExecutionSpan(UniversalBaseModel):
    name: typing.Literal["node.execution"] = "node.execution"
    events: typing.List[VellumNodeExecutionEvent]
    attributes: NodeExecutionSpanAttributes
    usage_result: typing.Optional[WorkflowExecutionUsageCalculationFulfilledBody] = None
    span_id: str
    start_ts: dt.datetime
    end_ts: dt.datetime
    parent_span_id: typing.Optional[str] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import types
from typing import TYPE_CHECKING, Generic, Optional, Type, TypeVar

from vellum.client import Vellum as VellumClient
from vellum.workflows.types.generics import WorkflowType
from vellum_ee.workflows.display.types import WorkflowDisplayContext
from vellum_ee.workflows.display.utils.registry import get_from_workflow_display_registry

if TYPE_CHECKING:
    from vellum_ee.workflows.display.workflows import BaseWorkflowDisplay


def _get_workflow_display_class(*, workflow_class: Type[WorkflowType]) -> Type["BaseWorkflowDisplay"]:
    workflow_display_class = get_from_workflow_display_registry(workflow_class)
    if workflow_display_class:
        return workflow_display_class

    base_workflow_display_class = _get_workflow_display_class(
        workflow_class=workflow_class.__bases__[0],
    )

    # mypy gets upset at dynamic TypeVar's, but it's technically allowed by python
    _WorkflowClassType = TypeVar(f"_{workflow_class.__name__}Type", bound=workflow_class)  # type: ignore[misc]
    # `base_workflow_display_class` is always a Generic class, so it's safe to index into it
    WorkflowDisplayBaseClass = base_workflow_display_class[_WorkflowClassType]  # type: ignore[index]

    WorkflowDisplayClass = types.new_class(
        f"{workflow_class.__name__}Display",
        bases=(WorkflowDisplayBaseClass, Generic[_WorkflowClassType]),
    )

    return WorkflowDisplayClass


def get_workflow_display(
    *,
    workflow_class: Type[WorkflowType],
    parent_display_context: Optional[WorkflowDisplayContext] = None,
    client: Optional[VellumClient] = None,
    dry_run: bool = False,
    # DEPRECATED: The following arguments will be removed in 0.15.0
    root_workflow_class: Optional[Type[WorkflowType]] = None,
    base_display_class: Optional[Type["BaseWorkflowDisplay"]] = None,
) -> "BaseWorkflowDisplay":
    return _get_workflow_display_class(workflow_class=workflow_class)(
        parent_display_context=parent_display_context,
        client=client,
        dry_run=dry_run,
    )



import pytest

from vellum.workflows.nodes import BaseNode
from vellum.workflows.references.lazy import LazyReference


@pytest.fixture
def mock_inspect_getsource(mocker):
    return mocker.patch("inspect.getsource")


@pytest.fixture
def mock_logger(mocker):
    return mocker.patch("vellum.workflows.references.lazy.logger")


def test_lazy_reference__inspect_getsource_fails(mock_inspect_getsource, mock_logger):
    # GIVEN getsource fails to resolve the lambda's source code
    mock_inspect_getsource.side_effect = Exception("test")

    # WHEN a node with a lazy reference is defined
    class MyNode(BaseNode):
        lazy_reference = LazyReference(lambda: MyNode.Outputs.foo)

    # THEN the name is the lambda function's name
    assert MyNode.lazy_reference.instance
    assert MyNode.lazy_reference.instance.name == "<lambda>"

    # AND sentry is notified
    assert mock_logger.exception.call_count == 1



import pytest
from dataclasses import dataclass
from enum import Enum
from unittest.mock import Mock
from typing import Annotated, Dict, List, Literal, Optional, Tuple, Union

from pydantic import BaseModel

from vellum.client.types.function_definition import FunctionDefinition
from vellum.client.types.string_vellum_value import StringVellumValue
from vellum.client.types.vellum_variable import VellumVariable
from vellum.workflows import BaseWorkflow
from vellum.workflows.inputs.base import BaseInputs
from vellum.workflows.nodes.bases.base import BaseNode
from vellum.workflows.state.base import BaseState
from vellum.workflows.utils.functions import (
    compile_function_definition,
    compile_inline_workflow_function_definition,
    compile_workflow_deployment_function_definition,
)


def test_compile_function_definition__just_name():
    # GIVEN a function with just a name
    def my_function():
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={"type": "object", "properties": {}, "required": []},
    )


def test_compile_function_definition__docstring():
    # GIVEN a function with a docstring
    def my_function():
        """This is a test function"""
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        description="This is a test function",
        parameters={"type": "object", "properties": {}, "required": []},
    )


def test_compile_function_definition__all_args():
    # GIVEN a function with args of all base types
    def my_function(a: str, b: int, c: float, d: bool, e: list, f: dict):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {
                "a": {"type": "string"},
                "b": {"type": "integer"},
                "c": {"type": "number"},
                "d": {"type": "boolean"},
                "e": {"type": "array"},
                "f": {"type": "object"},
            },
            "required": ["a", "b", "c", "d", "e", "f"],
        },
    )


def test_compile_function_definition__unions():
    # GIVEN a function with a union arg
    def my_function(a: Union[str, int]):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {
                "a": {"anyOf": [{"type": "string"}, {"type": "integer"}]},
            },
            "required": ["a"],
        },
    )


def test_compile_function_definition__optionals():
    # GIVEN a function with various ways to specify optionals
    def my_function(
        a: str,
        b: Optional[str],
        c: None,
        d: str = "hello",
        e: Optional[str] = None,
    ):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {
                "a": {"type": "string"},
                "b": {"anyOf": [{"type": "string"}, {"type": "null"}]},
                "c": {"type": "null"},
                "d": {"type": "string", "default": "hello"},
                "e": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None},
            },
            "required": ["a", "b", "c"],
        },
    )


def test_compile_function_definition__parameterized_dicts():
    # GIVEN a function with a parameterized dict
    def my_function(a: Dict[str, int]):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {
                "a": {"type": "object", "additionalProperties": {"type": "integer"}},
            },
            "required": ["a"],
        },
    )


def test_compile_function_definition__parameterized_lists():
    # GIVEN a function with a parameterized list
    def my_function(a: List[int]):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {
                "a": {"type": "array", "items": {"type": "integer"}},
            },
            "required": ["a"],
        },
    )


def test_compile_function_definition__dataclasses():
    # GIVEN a function with a dataclass
    @dataclass
    class MyDataClass:
        a: int
        b: str

    def my_function(c: MyDataClass):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {"c": {"$ref": "#/$defs/MyDataClass"}},
            "required": ["c"],
            "$defs": {
                "MyDataClass": {
                    "type": "object",
                    "properties": {"a": {"type": "integer"}, "b": {"type": "string"}},
                    "required": ["a", "b"],
                }
            },
        },
    )


def test_compile_function_definition__pydantic():
    # GIVEN a function with a pydantic model
    class MyPydanticModel(BaseModel):
        a: int
        b: str

    def my_function(c: MyPydanticModel):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {"c": {"$ref": "#/$defs/MyPydanticModel"}},
            "required": ["c"],
            "$defs": {
                "MyPydanticModel": {
                    "type": "object",
                    "properties": {"a": {"type": "integer"}, "b": {"type": "string"}},
                    "required": ["a", "b"],
                }
            },
        },
    )


def test_compile_function_definition__default_dataclass():
    # GIVEN a function with a dataclass
    @dataclass
    class MyDataClass:
        a: int
        b: str

    def my_function(c: MyDataClass = MyDataClass(a=1, b="hello")):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {"c": {"$ref": "#/$defs/MyDataClass", "default": {"a": 1, "b": "hello"}}},
            "required": [],
            "$defs": {
                "MyDataClass": {
                    "type": "object",
                    "properties": {"a": {"type": "integer"}, "b": {"type": "string"}},
                    "required": ["a", "b"],
                }
            },
        },
    )


def test_compile_function_definition__default_pydantic():
    # GIVEN a function with a pydantic model as the default value
    class MyPydanticModel(BaseModel):
        a: int
        b: str

    def my_function(c: MyPydanticModel = MyPydanticModel(a=1, b="hello")):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {"c": {"$ref": "#/$defs/MyPydanticModel", "default": {"a": 1, "b": "hello"}}},
            "required": [],
            "$defs": {
                "MyPydanticModel": {
                    "type": "object",
                    "properties": {"a": {"type": "integer"}, "b": {"type": "string"}},
                    "required": ["a", "b"],
                }
            },
        },
    )


def test_compile_function_definition__lambda():
    # GIVEN a lambda
    lambda_function = lambda x: x + 1  # noqa: E731

    # WHEN compiling the function
    compiled_function = compile_function_definition(lambda_function)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="<lambda>",
        parameters={"type": "object", "properties": {"x": {"type": "null"}}, "required": ["x"]},
    )


def test_compile_inline_workflow_function_definition():
    class MyNode(BaseNode):
        pass

    class MyWorkflow(BaseWorkflow):
        graph = MyNode

    # WHEN compiling the function
    compiled_function = compile_inline_workflow_function_definition(MyWorkflow)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_workflow",
        parameters={"type": "object", "properties": {}, "required": []},
    )


def test_compile_inline_workflow_function_definition__docstring():
    class MyNode(BaseNode):
        pass

    class MyWorkflow(BaseWorkflow):
        """
        This is a test workflow
        """

        graph = MyNode

    # WHEN compiling the function
    compiled_function = compile_inline_workflow_function_definition(MyWorkflow)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_workflow",
        description="\n        This is a test workflow\n        ",
        parameters={"type": "object", "properties": {}, "required": []},
    )


def test_compile_inline_workflow_function_definition__all_args():
    class MyInputs(BaseInputs):
        a: str
        b: int
        c: float
        d: bool
        e: list
        f: dict

    class MyNode(BaseNode):
        pass

    class MyWorkflow(BaseWorkflow[MyInputs, BaseState]):
        graph = MyNode

    # WHEN compiling the workflow
    compiled_function = compile_inline_workflow_function_definition(MyWorkflow)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_workflow",
        parameters={
            "type": "object",
            "properties": {
                "a": {"type": "string"},
                "b": {"type": "integer"},
                "c": {"type": "number"},
                "d": {"type": "boolean"},
                "e": {"type": "array"},
                "f": {"type": "object"},
            },
            "required": ["a", "b", "c", "d", "e", "f"],
        },
    )


def test_compile_inline_workflow_function_definition__unions():
    # GIVEN a workflow with a union
    class MyInputs(BaseInputs):
        a: Union[str, int]

    class MyNode(BaseNode):
        pass

    class MyWorkflow(BaseWorkflow[MyInputs, BaseState]):
        graph = MyNode

    # WHEN compiling the workflow
    compiled_function = compile_inline_workflow_function_definition(MyWorkflow)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_workflow",
        parameters={
            "type": "object",
            "properties": {"a": {"anyOf": [{"type": "string"}, {"type": "integer"}]}},
            "required": ["a"],
        },
    )


def test_compile_inline_workflow_function_definition__optionals():
    class MyInputs(BaseInputs):
        a: str
        b: Optional[str]
        c: None
        d: str = "hello"
        e: Optional[str] = None

    class MyNode(BaseNode):
        pass

    class MyWorkflow(BaseWorkflow[MyInputs, BaseState]):
        graph = MyNode

    # WHEN compiling the workflow
    compiled_function = compile_inline_workflow_function_definition(MyWorkflow)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_workflow",
        parameters={
            "type": "object",
            "properties": {
                "a": {"type": "string"},
                "b": {"anyOf": [{"type": "string"}, {"type": "null"}]},
                "c": {"type": "null"},
                "d": {"type": "string", "default": "hello"},
                "e": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None},
            },
            "required": ["a", "b", "c"],
        },
    )


def test_compile_workflow_deployment_function_definition__just_name():
    # GIVEN a mock Vellum client and deployment
    mock_client = Mock()
    mock_release = Mock()
    mock_release.workflow_version.input_variables = []
    mock_release.description = "This is a test deployment"
    mock_client.workflow_deployments.retrieve_workflow_deployment_release.return_value = mock_release

    deployment_config = {"deployment": "my_deployment", "release_tag": "latest"}

    # WHEN compiling the deployment workflow function
    compiled_function = compile_workflow_deployment_function_definition(deployment_config, mock_client)

    # THEN it should return the compiled function definition (same structure as function test)
    assert compiled_function == FunctionDefinition(
        name="my_deployment",
        description="This is a test deployment",
        parameters={"type": "object", "properties": {}, "required": []},
    )


def test_compile_workflow_deployment_function_definition__all_args():
    # GIVEN a mock Vellum client and deployment
    mock_client = Mock()
    mock_release = Mock()

    mock_inputs = []
    for key, vellum_type in [
        ("a", "STRING"),
        ("b", "NUMBER"),
        ("c", "JSON"),
        ("d", "CHAT_HISTORY"),
        ("e", "SEARCH_RESULTS"),
        ("f", "ERROR"),
        ("g", "ARRAY"),
        ("h", "FUNCTION_CALL"),
        ("i", "IMAGE"),
        ("j", "AUDIO"),
        ("k", "DOCUMENT"),
        ("l", "NULL"),
    ]:
        mock_input = VellumVariable(
            id=f"input_{key}",
            key=key,
            type=vellum_type,
            required=True,
            default=None,
        )
        mock_inputs.append(mock_input)

    mock_release.workflow_version.input_variables = mock_inputs
    mock_release.description = "This is a test deployment"
    mock_client.workflow_deployments.retrieve_workflow_deployment_release.return_value = mock_release

    deployment_config = {"deployment": "my_deployment", "release_tag": "latest"}

    # WHEN compiling the deployment workflow function
    compiled_function = compile_workflow_deployment_function_definition(deployment_config, mock_client)

    # THEN it should return the compiled function definition
    assert compiled_function == FunctionDefinition(
        name="my_deployment",
        description="This is a test deployment",
        parameters={
            "type": "object",
            "properties": {
                "a": {"type": "string"},
                "b": {"type": "number"},
                "c": {"type": "object"},
                "d": {"type": "array"},
                "e": {"type": "array"},
                "f": {"type": "object"},
                "g": {"type": "array"},
                "h": {"type": "object"},
                "i": {"type": "object"},
                "j": {"type": "object"},
                "k": {"type": "object"},
                "l": {"type": "null"},
            },
            "required": ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l"],
        },
    )


def test_compile_workflow_deployment_function_definition__defaults():
    # GIVEN a mock Vellum client and deployment
    mock_client = Mock()
    mock_release = Mock()

    mock_inputs = []

    mock_input_no_default = VellumVariable(
        id="no_default",
        key="no_default",
        type="STRING",
        required=True,
        default=None,
    )
    mock_inputs.append(mock_input_no_default)

    mock_input_null_default = VellumVariable(
        id="null_default",
        key="null_default",
        type="STRING",
        required=False,
        default=StringVellumValue(value=None),
    )
    mock_inputs.append(mock_input_null_default)

    mock_input_actual_default = VellumVariable(
        id="actual_default",
        key="actual_default",
        type="STRING",
        required=False,
        default=StringVellumValue(value="hello world"),
    )
    mock_inputs.append(mock_input_actual_default)

    mock_release.workflow_version.input_variables = mock_inputs
    mock_release.description = "This is a test deployment"
    mock_client.workflow_deployments.retrieve_workflow_deployment_release.return_value = mock_release

    deployment_config = {"deployment": "my_deployment", "release_tag": "latest"}

    # WHEN compiling the deployment workflow function
    compiled_function = compile_workflow_deployment_function_definition(deployment_config, mock_client)

    # THEN it should return the compiled function definition with proper default handling
    assert compiled_function == FunctionDefinition(
        name="my_deployment",
        description="This is a test deployment",
        parameters={
            "type": "object",
            "properties": {
                "no_default": {"type": "string"},
                "null_default": {"type": "string", "default": None},
                "actual_default": {"type": "string", "default": "hello world"},
            },
            "required": ["no_default"],
        },
    )


@pytest.mark.parametrize(
    "annotation,expected_schema",
    [
        (Literal["a", "b"], {"type": "string", "enum": ["a", "b"]}),
        (Literal["a", 1], {"enum": ["a", 1]}),
    ],
)
def test_compile_function_definition__literal(annotation, expected_schema):
    def my_function(a: annotation):  # type: ignore
        pass

    compiled_function = compile_function_definition(my_function)
    assert isinstance(compiled_function.parameters, dict)
    assert compiled_function.parameters["properties"]["a"] == expected_schema


def test_compile_function_definition__literal_type_not_in_map():
    class MyEnum(Enum):
        FOO = "foo"
        BAR = "bar"

    def my_function(a: Literal[MyEnum.FOO, MyEnum.BAR]):
        pass

    compiled_function = compile_function_definition(my_function)
    assert isinstance(compiled_function.parameters, dict)
    assert compiled_function.parameters["properties"]["a"] == {"enum": [MyEnum.FOO, MyEnum.BAR]}


def test_compile_function_definition__annotated_descriptions():
    # GIVEN a function with annotated parameters that include descriptions
    def my_function(
        bar: Annotated[str, "My bar parameter"],
        other: Annotated[int, "My other parameter"],
        regular_param: str,
        optional_param: Annotated[bool, "Optional boolean parameter"] = True,
    ):
        """Test function with annotated parameters."""
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition with descriptions
    assert compiled_function == FunctionDefinition(
        name="my_function",
        description="Test function with annotated parameters.",
        parameters={
            "type": "object",
            "properties": {
                "bar": {"type": "string", "description": "My bar parameter"},
                "other": {"type": "integer", "description": "My other parameter"},
                "regular_param": {"type": "string"},
                "optional_param": {"type": "boolean", "description": "Optional boolean parameter", "default": True},
            },
            "required": ["bar", "other", "regular_param"],
        },
    )


def test_compile_function_definition__annotated_without_description():
    # GIVEN a function with annotated parameters but no description metadata
    def my_function(param: Annotated[str, None]):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition without description
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {
                "param": {"type": "string"},
            },
            "required": ["param"],
        },
    )


def test_compile_function_definition__annotated_complex_types():
    # GIVEN a function with annotated types
    def my_function(
        location: Annotated[Literal["New York", "Portland"], "The location you found"],
        items: Annotated[List[str], "List of string items"],
        config: Annotated[Dict[str, int], "Configuration mapping"],
    ):
        pass

    # WHEN compiling the function
    compiled_function = compile_function_definition(my_function)

    # THEN it should return the compiled function definition with descriptions for complex types
    assert compiled_function == FunctionDefinition(
        name="my_function",
        parameters={
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "enum": ["New York", "Portland"],
                    "description": "The location you found",
                },
                "items": {"type": "array", "items": {"type": "string"}, "description": "List of string items"},
                "config": {
                    "type": "object",
                    "additionalProperties": {"type": "integer"},
                    "description": "Configuration mapping",
                },
            },
            "required": ["location", "items", "config"],
        },
    )


@pytest.mark.parametrize(
    "annotation,expected_schema",
    [
        (
            Tuple[int, ...],
            {"type": "array", "items": {"type": "integer"}},
        ),
        (
            Tuple[int, str],
            {
                "type": "array",
                "prefixItems": [{"type": "integer"}, {"type": "string"}],
                "minItems": 2,
                "maxItems": 2,
            },
        ),
    ],
)
def test_compile_function_definition__tuples(annotation, expected_schema):
    def my_function(a: annotation):  # type: ignore
        pass

    compiled_function = compile_function_definition(my_function)
    assert isinstance(compiled_function.parameters, dict)
    assert compiled_function.parameters["properties"]["a"] == expected_schema



from uuid import UUID
from typing import Any, ClassVar, Dict, Generic, List, Optional, Union, cast

from vellum import ChatHistoryInput, ChatMessage, JsonInput, MetricDefinitionInput, NumberInput, StringInput
from vellum.client import ApiError
from vellum.client.core import RequestOptions
from vellum.workflows.constants import LATEST_RELEASE_TAG
from vellum.workflows.errors.types import WorkflowErrorCode
from vellum.workflows.exceptions import NodeException
from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.outputs.base import BaseOutputs
from vellum.workflows.types.core import EntityInputsInterface, MergeBehavior
from vellum.workflows.types.generics import StateType


class GuardrailNode(BaseNode[StateType], Generic[StateType]):
    """
    Used to execute a Metric Definition and surface a float output representing the score.

    metric_definition: Union[UUID, str] - Either the Metric Definition's UUID or its name.
    metric_inputs: EntityInputsInterface - The inputs for the Metric
    release_tag: str - The release tag to use for the Metric
    request_options: Optional[RequestOptions] - The request options to use for the Metric
    """

    metric_definition: ClassVar[Union[UUID, str]]

    metric_inputs: ClassVar[EntityInputsInterface] = {}
    release_tag: str = LATEST_RELEASE_TAG

    request_options: Optional[RequestOptions] = None

    class Trigger(BaseNode.Trigger):
        merge_behavior = MergeBehavior.AWAIT_ANY

    class Outputs(BaseOutputs):
        score: float
        normalized_score: Optional[float]
        log: Optional[str]
        reason: Optional[str]

    def run(self) -> Outputs:
        try:
            metric_execution = self._context.vellum_client.metric_definitions.execute_metric_definition(
                self.metric_definition if isinstance(self.metric_definition, str) else str(self.metric_definition),
                inputs=self._compile_metric_inputs(),
                release_tag=self.release_tag,
                request_options=self.request_options,
            )

        except ApiError:
            raise NodeException(
                code=WorkflowErrorCode.NODE_EXECUTION,
                message="Failed to execute metric definition",
            )

        metric_outputs = {output.name: output.value for output in metric_execution.outputs}
        SCORE_KEY = "score"
        NORMALIZED_SCORE_KEY = "normalized_score"
        LOG_KEY = "log"
        REASON_KEY = "reason"

        score = metric_outputs.get(SCORE_KEY)
        if not isinstance(score, float):
            raise NodeException(
                message=f"Metric execution must have one output named '{SCORE_KEY}' with type 'float'",
                code=WorkflowErrorCode.INVALID_OUTPUTS,
            )
        metric_outputs.pop(SCORE_KEY)

        if NORMALIZED_SCORE_KEY in metric_outputs:
            normalized_score = metric_outputs.pop(NORMALIZED_SCORE_KEY)
            if not isinstance(normalized_score, float):
                raise NodeException(
                    message=f"Metric execution must have one output named '{NORMALIZED_SCORE_KEY}' with type 'float'",
                    code=WorkflowErrorCode.INVALID_OUTPUTS,
                )
        else:
            normalized_score = None

        if LOG_KEY in metric_outputs:
            log = metric_outputs.pop(LOG_KEY) or ""
            if not isinstance(log, str):
                raise NodeException(
                    message="Metric execution log output must be of type 'str'",
                    code=WorkflowErrorCode.INVALID_OUTPUTS,
                )
        else:
            log = None

        if REASON_KEY in metric_outputs:
            reason = metric_outputs.pop(REASON_KEY) or ""
            if not isinstance(reason, str):
                raise NodeException(
                    message="Metric execution reason output must be of type 'str'",
                    code=WorkflowErrorCode.INVALID_OUTPUTS,
                )
        else:
            reason = None

        return self.Outputs(score=score, normalized_score=normalized_score, log=log, reason=reason, **metric_outputs)

    def _compile_metric_inputs(self) -> List[MetricDefinitionInput]:
        # TODO: We may want to consolidate with prompt deployment input compilation
        # https://app.shortcut.com/vellum/story/4117

        compiled_inputs: List[MetricDefinitionInput] = []

        for input_name, input_value in self.metric_inputs.items():
            if isinstance(input_value, str):
                compiled_inputs.append(
                    StringInput(
                        name=input_name,
                        value=input_value,
                    )
                )
            elif isinstance(input_value, list) and all(isinstance(message, ChatMessage) for message in input_value):
                compiled_inputs.append(
                    ChatHistoryInput(
                        name=input_name,
                        value=cast(List[ChatMessage], input_value),
                    )
                )
            elif isinstance(input_value, dict):
                compiled_inputs.append(
                    JsonInput(
                        name=input_name,
                        value=cast(Dict[str, Any], input_value),
                    )
                )
            elif isinstance(input_value, (int, float)):
                compiled_inputs.append(
                    NumberInput(
                        name=input_name,
                        value=input_value,
                    )
                )
            else:
                raise NodeException(
                    message=f"Unrecognized input type for input '{input_name}'",
                    code=WorkflowErrorCode.INVALID_INPUTS,
                )

        return compiled_inputs



import sys
from uuid import uuid4
from typing import Type, cast

from vellum.client.core.pydantic_utilities import UniversalBaseModel
from vellum.client.types.code_executor_response import CodeExecutorResponse
from vellum.client.types.number_vellum_value import NumberVellumValue
from vellum.workflows import BaseWorkflow
from vellum.workflows.nodes import BaseNode
from vellum.workflows.state.context import WorkflowContext
from vellum_ee.workflows.server.virtual_file_loader import VirtualFileFinder


def test_load_workflow_event_display_context():
    from vellum.workflows.events.workflow import WorkflowEventDisplayContext

    # We are actually just ensuring there are no circular dependencies when
    # our Workflow Server imports this class.
    assert issubclass(WorkflowEventDisplayContext, UniversalBaseModel)


def test_load_from_module__lazy_reference_in_file_loader():
    # GIVEN a workflow module with a node containing a lazy reference
    files = {
        "__init__.py": "",
        "workflow.py": """\
from vellum.workflows import BaseWorkflow
from .nodes.start_node import StartNode

class Workflow(BaseWorkflow):
    graph = StartNode
""",
        "nodes/__init__.py": """\
from .start_node import StartNode

__all__ = [
    "StartNode",
]
""",
        "nodes/start_node.py": """\
from vellum.workflows.nodes import BaseNode
from vellum.workflows.references import LazyReference

class StartNode(BaseNode):
    foo = LazyReference(lambda: StartNode.Outputs.bar)

    class Outputs(BaseNode.Outputs):
        bar = str
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    sys.meta_path.append(VirtualFileFinder(files, namespace))

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow()

    # THEN the workflow is successfully initialized
    assert workflow

    # AND the graph is just a BaseNode
    # ideally this would be true, but the loader uses a different BaseNode class definition than
    # the one in this test module.
    # assert isinstance(workflow.graph, BaseNode)
    start_node = cast(Type[BaseNode], workflow.graph)
    assert start_node.__bases__ == (BaseNode,)

    # AND the lazy reference has the correct name
    assert start_node.foo.instance
    assert start_node.foo.instance.name == "StartNode.Outputs.bar"


def test_load_from_module__ts_code_in_file_loader(
    vellum_client,
):
    # GIVEN typescript code
    ts_code = """async function main(): any {
  return 5;
}"""

    # AND a workflow module with only a code execution node
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from .nodes.code_execution_node import CodeExecutionNode

class Workflow(BaseWorkflow):
    graph = CodeExecutionNode

    class Outputs(BaseWorkflow.Outputs):
       final_output = CodeExecutionNode.Outputs.result
""",
        "nodes/__init__.py": """
from .code_execution_node import CodeExecutionNode

__all__ = ["CodeExecutionNode"]
""",
        "nodes/code_execution_node/__init__.py": """
from typing import Any

from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

class CodeExecutionNode(BaseCodeExecutionNode[BaseState, int]):
    filepath = "./script.ts"
    code_inputs = {}
    runtime = "TYPESCRIPT_5_3_3"
    packages = []
""",
        "nodes/code_execution_node/script.ts": ts_code,
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=5),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    event = workflow.run()
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the code execution result
    assert event.body.outputs == {"final_output": 5.0}


def test_load_from_module__simple_code_execution_node_with_try(
    vellum_client,
):
    # GIVEN a simple Python script
    py_code = """def main() -> int:
    print("Hello")
    return 1
"""

    # AND a workflow module with only a code execution node
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.code_execution_node import CodeExecutionNode

class Workflow(BaseWorkflow):
    graph = CodeExecutionNode

    class Outputs(BaseWorkflow.Outputs):
       final_output = CodeExecutionNode.Outputs.result
""",
        "nodes/__init__.py": """
from .code_execution_node import CodeExecutionNode

__all__ = ["CodeExecutionNode"]
""",
        "nodes/code_execution_node/__init__.py": """
from typing import Union

from vellum.workflows.nodes.core import TryNode
from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

@TryNode.wrap()
class CodeExecutionNode(BaseCodeExecutionNode[BaseState, Union[float, int]]):
    filepath = "./script.py"
    code_inputs = {}
    runtime = "PYTHON_3_11_6"
    packages = []
""",
        "nodes/code_execution_node/script.py": py_code,
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=1),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    # WHEN we run the workflow
    event = workflow.run()

    # THEN the execution is fulfilled
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the code execution result
    assert event.body.outputs == {"final_output": 1.0}


def test_load_from_module__simple_code_execution_node_with_retry(
    vellum_client,
):
    # GIVEN a simple Python script
    py_code = """def main() -> int:
    print("Hello")
    return 1
"""

    # AND a workflow module with only a code execution node wrapped with RetryNode
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.code_execution_node import CodeExecutionNode

class Workflow(BaseWorkflow):
    graph = CodeExecutionNode

    class Outputs(BaseWorkflow.Outputs):
       final_output = CodeExecutionNode.Outputs.result
""",
        "nodes/__init__.py": """
from .code_execution_node import CodeExecutionNode

__all__ = ["CodeExecutionNode"]
""",
        "nodes/code_execution_node/__init__.py": """
from typing import Union

from vellum.workflows.nodes.core.retry_node.node import RetryNode
from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

@RetryNode.wrap(max_attempts=3)
class CodeExecutionNode(BaseCodeExecutionNode[BaseState, Union[float, int]]):
    filepath = "./script.py"
    code_inputs = {}
    runtime = "PYTHON_3_11_6"
    packages = []
""",
        "nodes/code_execution_node/script.py": py_code,
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=1),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    # WHEN we run the workflow
    event = workflow.run()

    # THEN the execution is fulfilled
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the code execution result
    assert event.body.outputs == {"final_output": 1.0}


def test_load_from_module__code_execution_within_subworkflow(
    vellum_client,
):
    # GIVEN a simple Python script
    py_code = """def main() -> int:
    print("Hello")
    return 1
"""

    # AND a workflow module with a subworkflow containing a code execution node
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.subworkflow import Subworkflow

class Workflow(BaseWorkflow):
    graph = Subworkflow

    class Outputs(BaseWorkflow.Outputs):
       final_output = Subworkflow.Outputs.result
""",
        "nodes/__init__.py": """
from .subworkflow import Subworkflow

__all__ = ["Subworkflow"]
""",
        "nodes/subworkflow/__init__.py": """
from vellum.workflows.nodes.displayable import InlineSubworkflowNode

from .workflow import SubworkflowWorkflow

class Subworkflow(InlineSubworkflowNode):
    subworkflow = SubworkflowWorkflow
""",
        "nodes/subworkflow/nodes/__init__.py": """
from .code_execution_node import CodeExecutionNode

__all__ = ["CodeExecutionNode"]
""",
        "nodes/subworkflow/nodes/code_execution_node/__init__.py": """
from typing import Union

from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

class CodeExecutionNode(BaseCodeExecutionNode[BaseState, Union[float, int]]):
    filepath = "./script.py"
    code_inputs = {}
    runtime = "PYTHON_3_11_6"
    packages = []
""",
        "nodes/subworkflow/nodes/code_execution_node/script.py": py_code,
        "nodes/subworkflow/workflow.py": """
from vellum.workflows import BaseWorkflow

from .nodes.code_execution_node import CodeExecutionNode

class SubworkflowWorkflow(BaseWorkflow):
    graph = CodeExecutionNode

    class Outputs(BaseWorkflow.Outputs):
        result = CodeExecutionNode.Outputs.result
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=1),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    # WHEN we run the workflow
    event = workflow.run()

    # THEN the execution is fulfilled
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the code execution result from the subworkflow
    assert event.body.outputs == {"final_output": 1.0}


def test_load_from_module__code_execution_within_map_node(
    vellum_client,
):
    # GIVEN a simple Python script
    py_code = """def main() -> int:
    print("Hello")
    return 1
"""

    # AND a workflow module with a map node containing a code execution node
    files = {
        "__init__.py": "",
        "workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.map_node import MapNode

class Workflow(BaseWorkflow):
    graph = MapNode
    
    class Outputs(BaseWorkflow.Outputs):  # noqa: W293
       results = MapNode.Outputs.final_output
""",
        "nodes/__init__.py": """
from .map_node import MapNode

__all__ = ["MapNode"]
""",
        "nodes/map_node/__init__.py": """
from vellum.workflows.nodes.core.map_node import MapNode as BaseMapNode

from .workflow import MapNodeWorkflow

class MapNode(BaseMapNode):
    items = ["foo", "bar", "baz"]
    subworkflow = MapNodeWorkflow
    max_concurrency = 4
""",
        "nodes/map_node/nodes/__init__.py": """
from .code_execution_node import CodeExecutionNode

__all__ = ["CodeExecutionNode"]
""",
        "nodes/map_node/nodes/code_execution_node/__init__.py": """
from typing import Union

from vellum.workflows.nodes.displayable import CodeExecutionNode as BaseCodeExecutionNode
from vellum.workflows.state import BaseState

class CodeExecutionNode(BaseCodeExecutionNode[BaseState, Union[float, int]]):
    filepath = "./script.py"
    code_inputs = {}
    runtime = "PYTHON_3_11_6"
    packages = []
""",
        "nodes/map_node/nodes/code_execution_node/script.py": py_code,
        "nodes/map_node/workflow.py": """
from vellum.workflows import BaseWorkflow
from vellum.workflows.state import BaseState

from .nodes.code_execution_node import CodeExecutionNode

class MapNodeInputs:
    item: str
    index: int

class MapNodeWorkflow(BaseWorkflow):
    graph = CodeExecutionNode
    
    class Outputs(BaseWorkflow.Outputs):  # noqa: W293
        final_output = CodeExecutionNode.Outputs.result
""",
    }

    namespace = str(uuid4())

    # AND the virtual file loader is registered
    finder = VirtualFileFinder(files, namespace)
    sys.meta_path.append(finder)

    # AND we know what the Code Execution Node will respond with
    mock_code_execution = CodeExecutorResponse(
        log="hello",
        output=NumberVellumValue(value=3),
    )
    vellum_client.execute_code.return_value = mock_code_execution

    # WHEN the workflow is loaded
    Workflow = BaseWorkflow.load_from_module(namespace)
    workflow = Workflow(context=WorkflowContext(generated_files=files))

    # THEN the workflow is successfully initialized
    assert workflow

    # WHEN we run the workflow
    event = workflow.run()

    # THEN the execution is fulfilled
    assert event.name == "workflow.execution.fulfilled"

    # AND we get the map node results as a list
    assert event.body.outputs == {"results": [1.0, 1.0, 1.0]}



# This file was auto-generated by Fern from our API Definition.

import typing
from .templating_node_string_result import TemplatingNodeStringResult
from .templating_node_number_result import TemplatingNodeNumberResult
from .templating_node_json_result import TemplatingNodeJsonResult
from .templating_node_chat_history_result import TemplatingNodeChatHistoryResult
from .templating_node_search_results_result import TemplatingNodeSearchResultsResult
from .templating_node_error_result import TemplatingNodeErrorResult
from .templating_node_array_result import TemplatingNodeArrayResult
from .templating_node_function_call_result import TemplatingNodeFunctionCallResult

TemplatingNodeResultOutput = typing.Union[
    TemplatingNodeStringResult,
    TemplatingNodeNumberResult,
    TemplatingNodeJsonResult,
    TemplatingNodeChatHistoryResult,
    TemplatingNodeSearchResultsResult,
    TemplatingNodeErrorResult,
    TemplatingNodeArrayResult,
    TemplatingNodeFunctionCallResult,
]



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations
import typing
from .vellum_value_logical_condition_request import VellumValueLogicalConditionRequest
import typing

if typing.TYPE_CHECKING:
    from .vellum_value_logical_condition_group_request import VellumValueLogicalConditionGroupRequest
VellumValueLogicalExpressionRequest = typing.Union[
    VellumValueLogicalConditionRequest, "VellumValueLogicalConditionGroupRequest"
]



import pytest
from uuid import UUID

from vellum.workflows.nodes.bases import BaseNode
from vellum.workflows.ports.port import Port
from vellum.workflows.references.constant import ConstantValueReference
from vellum_ee.workflows.display.nodes.base_node_display import BaseNodeDisplay
from vellum_ee.workflows.display.nodes.get_node_display_class import get_node_display_class
from vellum_ee.workflows.display.types import WorkflowDisplayContext


@pytest.fixture
def node_with_implicit_properties():
    class MyNode(BaseNode):
        pass

    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        pass

    expected_id = UUID("ace7f746-4fe6-45c7-8207-fc8a4d0c7f6f")

    return MyNodeDisplay, expected_id


@pytest.fixture
def node_with_explicit_properties():
    explicit_id = UUID("a422f67a-1d37-43f0-bdfc-1e4618c9496d")

    class MyNode(BaseNode):
        pass

    class MyNodeDisplay(BaseNodeDisplay[MyNode]):
        node_id = explicit_id

    return MyNodeDisplay, explicit_id


@pytest.fixture(
    params=[
        "node_with_implicit_properties",
        "node_with_explicit_properties",
    ]
)
def node_info(request):
    return request.getfixturevalue(request.param)


def test_get_id(node_info):
    node_display, expected_id = node_info

    assert node_display().node_id == expected_id
    assert node_display.infer_node_class().__id__ == expected_id


def test_serialize_condition__accessor_expression():
    # GIVEN a node with an accessor expression in a Port
    class MyNode(BaseNode):
        class Ports(BaseNode.Ports):
            foo = Port.on_if(ConstantValueReference({"hello": "world"})["hello"])

    # WHEN we serialize the node
    node_display_class = get_node_display_class(MyNode)
    data = node_display_class().serialize(WorkflowDisplayContext())

    # THEN the condition should be serialized correctly
    assert data["ports"] == [
        {
            "id": "7de6ea94-7f6c-475e-8f38-ec8ac317fd19",
            "name": "foo",
            "type": "IF",
            "expression": {
                "type": "BINARY_EXPRESSION",
                "lhs": {
                    "type": "CONSTANT_VALUE",
                    "value": {
                        "type": "JSON",
                        "value": {"hello": "world"},
                    },
                },
                "operator": "accessField",
                "rhs": {"type": "CONSTANT_VALUE", "value": {"type": "STRING", "value": "hello"}},
            },
        }
    ]


def test_serialize_display_data():
    # GIVEN a node with an accessor expression in a Port
    class MyNode(BaseNode):
        """I hope this works"""

        pass

    # WHEN we serialize the node
    node_display_class = get_node_display_class(MyNode)
    data = node_display_class().serialize(WorkflowDisplayContext())

    # THEN the condition should be serialized correctly
    assert data["display_data"] == {
        "position": {"x": 0.0, "y": 0.0},
        "comment": {"value": "I hope this works"},
    }



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .api_request_parent_context import ApiRequestParentContext
from .external_parent_context import ExternalParentContext
from .node_parent_context import NodeParentContext
from .prompt_deployment_parent_context import PromptDeploymentParentContext
from .span_link import SpanLink
from .workflow_deployment_parent_context import WorkflowDeploymentParentContext
from .workflow_parent_context import WorkflowParentContext
from .workflow_sandbox_parent_context import WorkflowSandboxParentContext
import typing
from .parent_context import ParentContext
from .node_execution_rejected_body import NodeExecutionRejectedBody
import datetime as dt
from .api_version_enum import ApiVersionEnum
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic


class NodeExecutionRejectedEvent(UniversalBaseModel):
    parent: typing.Optional[ParentContext] = None
    links: typing.Optional[typing.List[SpanLink]] = None
    name: typing.Literal["node.execution.rejected"] = "node.execution.rejected"
    body: NodeExecutionRejectedBody
    id: str
    timestamp: dt.datetime
    api_version: typing.Optional[ApiVersionEnum] = None
    trace_id: str
    span_id: str

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations
import typing
from .string_vellum_value import StringVellumValue
from .number_vellum_value import NumberVellumValue
from .json_vellum_value import JsonVellumValue
from .audio_vellum_value import AudioVellumValue
from .video_vellum_value import VideoVellumValue
from .image_vellum_value import ImageVellumValue
from .document_vellum_value import DocumentVellumValue
from .function_call_vellum_value import FunctionCallVellumValue
from .error_vellum_value import ErrorVellumValue
from .chat_history_vellum_value import ChatHistoryVellumValue
from .search_results_vellum_value import SearchResultsVellumValue
from .thinking_vellum_value import ThinkingVellumValue
import typing

if typing.TYPE_CHECKING:
    from .array_vellum_value import ArrayVellumValue
VellumValue = typing.Union[
    StringVellumValue,
    NumberVellumValue,
    JsonVellumValue,
    AudioVellumValue,
    VideoVellumValue,
    ImageVellumValue,
    DocumentVellumValue,
    FunctionCallVellumValue,
    ErrorVellumValue,
    "ArrayVellumValue",
    ChatHistoryVellumValue,
    SearchResultsVellumValue,
    ThinkingVellumValue,
]



# This file was auto-generated by Fern from our API Definition.

from __future__ import annotations
from ..core.pydantic_utilities import UniversalBaseModel
from .array_vellum_value_request import ArrayVellumValueRequest
import typing
from .condition_combinator import ConditionCombinator
from ..core.pydantic_utilities import IS_PYDANTIC_V2
import pydantic
from ..core.pydantic_utilities import update_forward_refs


class VellumValueLogicalConditionGroupRequest(UniversalBaseModel):
    """
    A higher-order condition that combines one or more basic conditions or other higher-order conditions.
    """

    type: typing.Literal["LOGICAL_CONDITION_GROUP"] = "LOGICAL_CONDITION_GROUP"
    conditions: typing.List["VellumValueLogicalExpressionRequest"]
    combinator: ConditionCombinator
    negated: bool

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow


from .vellum_value_logical_expression_request import VellumValueLogicalExpressionRequest  # noqa: E402

update_forward_refs(VellumValueLogicalConditionGroupRequest)



# This file was auto-generated by Fern from our API Definition.

from ..core.pydantic_utilities import UniversalBaseModel
from .array_vellum_value import ArrayVellumValue
import pydantic
import typing
from .entity_status import EntityStatus
from .environment_enum import EnvironmentEnum
import datetime as dt
from .vellum_variable import VellumVariable
from ..core.pydantic_utilities import IS_PYDANTIC_V2


class SlimWorkflowDeployment(UniversalBaseModel):
    id: str
    name: str = pydantic.Field()
    """
    A name that uniquely identifies this workflow deployment within its workspace
    """

    label: str = pydantic.Field()
    """
    A human-readable label for the workflow deployment
    """

    status: typing.Optional[EntityStatus] = pydantic.Field(default=None)
    """
    The current status of the workflow deployment
    
    * `ACTIVE` - Active
    * `ARCHIVED` - Archived
    """

    environment: typing.Optional[EnvironmentEnum] = pydantic.Field(default=None)
    """
    The environment this workflow deployment is used in
    
    * `DEVELOPMENT` - Development
    * `STAGING` - Staging
    * `PRODUCTION` - Production
    """

    created: dt.datetime
    last_deployed_on: dt.datetime
    input_variables: typing.List[VellumVariable] = pydantic.Field()
    """
    The input variables this Workflow Deployment expects to receive values for when it is executed.
    """

    output_variables: typing.List[VellumVariable] = pydantic.Field()
    """
    The output variables this Workflow Deployment will produce when it is executed.
    """

    description: typing.Optional[str] = pydantic.Field(default=None)
    """
    A human-readable description of the workflow deployment
    """

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow
